# This code is part of Qiskit.
#
# (C) Copyright IBM 2021
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Driver for a synthesis routine which emits optimal XX-based circuits.
"""

import heapq
import math
from operator import itemgetter
from typing import Callable, Optional

import numpy as np

from qiskit.circuit.quantumcircuit import QuantumCircuit
from qiskit.circuit.library.standard_gates import RXXGate, RZXGate
from qiskit.exceptions import QiskitError
from qiskit.quantum_info.operators import Operator
from qiskit.quantum_info.synthesis.one_qubit_decompose import ONE_QUBIT_EULER_BASIS_GATES
from qiskit.quantum_info.synthesis.two_qubit_decompose import TwoQubitWeylDecomposition

from .circuits import apply_reflection, apply_shift, canonical_xx_circuit
from .utilities import EPSILON
from .polytopes import XXPolytope


def _average_infidelity(p, q):
    """
    Computes the infidelity distance between two points p, q expressed in positive canonical
    coordinates.
    """

    a0, b0, c0 = p
    a1, b1, c1 = q

    return 1 - 1 / 20 * (
        4
        + 16
        * (
            math.cos(a0 - a1) ** 2 * math.cos(b0 - b1) ** 2 * math.cos(c0 - c1) ** 2
            + math.sin(a0 - a1) ** 2 * math.sin(b0 - b1) ** 2 * math.sin(c0 - c1) ** 2
        )
    )


class XXDecomposer:
    """
    A class for optimal decomposition of 2-qubit unitaries into 2-qubit basis gates of XX type
    (i.e., each locally equivalent to CAN(alpha, 0, 0) for a possibly varying alpha).

    Args:
        euler_basis: Basis string provided to OneQubitEulerDecomposer for 1Q synthesis.
            Defaults to "U".
        embodiments: A dictionary mapping interaction strengths alpha to native circuits which
            embody the gate CAN(alpha, 0, 0). Strengths are taken so that pi/2 represents the class
            of a full CX.
        backup_optimizer: If supplied, defers synthesis to this callable when XXDecomposer
            has no efficient decomposition of its own. Useful for special cases involving 2 or 3
            applications of XX(pi/2), in which case standard synthesis methods provide lower
            1Q gate count.

    .. note::
        If ``embodiments`` is not passed, or if an entry is missing, it will be populated as needed
        using the method ``_default_embodiment``.
    """

    def __init__(
        self,
        euler_basis: str = "U",
        embodiments: Optional[dict] = None,
        backup_optimizer: Optional[Callable] = None,
    ):
        from qiskit.transpiler.passes.optimization.optimize_1q_decomposition import (
            Optimize1qGatesDecomposition,  # pylint: disable=cyclic-import
        )

        self._decomposer1q = Optimize1qGatesDecomposition(ONE_QUBIT_EULER_BASIS_GATES[euler_basis])
        self.gate = RZXGate(np.pi / 2)
        self.embodiments = embodiments if embodiments is not None else {}
        self.backup_optimizer = backup_optimizer

        self._check_embodiments()

    @staticmethod
    def _default_embodiment(strength):
        """
        If the user does not provide a custom implementation of XX(strength), then this routine
        defines a default implementation using RZX.
        """
        xx_circuit = QuantumCircuit(2)

        # NOTE: One could branch here on `strength == np.pi / 2` and decide to use a CX-based
        #       circuit in this one case where it's available.
        xx_circuit.h(0)
        xx_circuit.rzx(strength, 0, 1)
        xx_circuit.h(0)

        return xx_circuit

    def _check_embodiments(self):
        """
        Checks that `self.embodiments` is populated with legal circuit embodiments: the key-value
        pair (angle, circuit) satisfies Operator(circuit) approx RXX(angle).to_matrix().
        """
        # pylint: disable=cyclic-import
        from qiskit.quantum_info.operators.measures import average_gate_fidelity

        for angle, embodiment in self.embodiments.items():
            actual = Operator(RXXGate(angle))
            purported = Operator(embodiment)
            if average_gate_fidelity(actual, purported) < 1 - EPSILON:
                raise QiskitError(
                    f"RXX embodiment provided for angle {angle} disagrees with RXXGate({angle})"
                )

    @staticmethod
    def _best_decomposition(canonical_coordinate, available_strengths):
        """
        Finds the cheapest sequence of `available_strengths` which supports the best approximation
        to `canonical_coordinate`. Returns a dictionary with keys "cost", "point", and "operations".

        NOTE: `canonical_coordinate` is a positive canonical coordinate. `strengths` is a dictionary
              mapping the available strengths to their (infidelity) costs, with the strengths
              themselves normalized so that pi/2 represents CX = RZX(pi/2).
        """
        best_point, best_cost, best_sequence = [0, 0, 0], 1.0, []
        priority_queue = []
        heapq.heappush(priority_queue, (0, []))
        canonical_coordinate = np.array(canonical_coordinate)

        while True:
            sequence_cost, sequence = heapq.heappop(priority_queue)

            strength_polytope = XXPolytope.from_strengths(*[x / 2 for x in sequence])
            candidate_point = strength_polytope.nearest(canonical_coordinate)
            candidate_cost = sequence_cost + _average_infidelity(
                canonical_coordinate, candidate_point
            )

            if candidate_cost < best_cost:
                best_point, best_cost, best_sequence = candidate_point, candidate_cost, sequence

            if strength_polytope.member(canonical_coordinate):
                break

            for strength, extra_cost in available_strengths.items():
                if len(sequence) == 0 or strength <= sequence[-1]:
                    heapq.heappush(
                        priority_queue, (sequence_cost + extra_cost, sequence + [strength])
                    )

        return {"point": best_point, "cost": best_cost, "sequence": best_sequence}

    def num_basis_gates(self, unitary):
        """
        Counts the number of gates that would be emitted during re-synthesis.

        NOTE: Used by ConsolidateBlocks.
        """
        strengths = self._strength_to_infidelity(1.0)

        # get the associated _positive_ canonical coordinate
        weyl_decomposition = TwoQubitWeylDecomposition(unitary)
        target = [getattr(weyl_decomposition, x) for x in ("a", "b", "c")]
        if target[-1] < -EPSILON:
            target = [np.pi / 2 - target[0], target[1], -target[2]]

        best_sequence = self._best_decomposition(target, strengths)["sequence"]
        return len(best_sequence)

    @staticmethod
    def _strength_to_infidelity(basis_fidelity, approximate=False):
        """
        Converts a dictionary mapping ZX strengths to fidelities to a dictionary mapping ZX
        strengths to infidelities. Also supports one of the other formats Qiskit uses: if only a
        lone float is supplied, it extends it from CX over CX/2 and CX/3 by linear decay.
        """

        if isinstance(basis_fidelity, float):
            if not approximate:
                slope, offset = 1e-10, 1e-12
            else:
                slope, offset = (1 - basis_fidelity) / 2, (1 - basis_fidelity) / 2
            return {
                strength: slope * strength / (np.pi / 2) + offset
                for strength in [np.pi / 2, np.pi / 4, np.pi / 6]
            }
        elif isinstance(basis_fidelity, dict):
            return {
                strength: (1 - fidelity if approximate else 1e-12 + 1e-10 * strength / (np.pi / 2))
                for (strength, fidelity) in basis_fidelity.items()
            }

        raise TypeError("Unknown basis_fidelity payload.")

    def __call__(self, unitary, basis_fidelity=1.0, approximate=True):
        """
        Fashions a circuit which (perhaps `approximate`ly) models the special unitary operation
        `unitary`, using the circuit templates supplied at initialization as `embodiments`.  The
        routine uses `basis_fidelity` to select the optimal circuit template, including when
        performing exact synthesis; the contents of `basis_fidelity` is a dictionary mapping
        interaction strengths (scaled so that CX = RZX(pi/2) corresponds to pi/2) to circuit
        fidelities.

        Args:
            unitary (Operator or ndarray): 4x4 unitary to synthesize.
            basis_fidelity (dict or float): Fidelity of basis gates. Can be either (1) a dictionary
                mapping XX angle values to fidelity at that angle; or (2) a single float f,
                interpreted as {pi: f, pi/2: f/2, pi/3: f/3} .
            approximate (bool): Approximates if basis fidelities are less than 1.0 .
        Returns:
            QuantumCircuit: Synthesized circuit.
        """
        strength_to_infidelity = self._strength_to_infidelity(
            basis_fidelity, approximate=approximate
        )
        from qiskit.extensions import UnitaryGate  # pylint: disable=cyclic-import

        # get the associated _positive_ canonical coordinate
        weyl_decomposition = TwoQubitWeylDecomposition(unitary)
        target = [getattr(weyl_decomposition, x) for x in ("a", "b", "c")]
        if target[-1] < -EPSILON:
            target = [np.pi / 2 - target[0], target[1], -target[2]]

        # scan for the best point
        best_point, best_sequence = itemgetter("point", "sequence")(
            self._best_decomposition(target, strength_to_infidelity)
        )
        # build the circuit building this canonical gate
        embodiments = {
            k: self.embodiments.get(k, self._default_embodiment(k))
            for k, v in strength_to_infidelity.items()
        }
        circuit = canonical_xx_circuit(best_point, best_sequence, embodiments)

        if best_sequence == [np.pi / 2, np.pi / 2, np.pi / 2] and self.backup_optimizer is not None:
            return self.backup_optimizer(unitary, basis_fidelity=basis_fidelity)

        # change to positive canonical coordinates
        if weyl_decomposition.c >= -EPSILON:
            # if they're the same...
            corrected_circuit = QuantumCircuit(2)
            corrected_circuit.rz(np.pi, [0])
            corrected_circuit.compose(circuit, [0, 1], inplace=True)
            corrected_circuit.rz(-np.pi, [0])
            circuit = corrected_circuit
        else:
            # else they're in the "positive" scissors part...
            corrected_circuit = QuantumCircuit(2)
            _, source_reflection, _ = apply_reflection("reflect XX, ZZ", [0, 0, 0])
            _, source_shift, _ = apply_shift("X shift", [0, 0, 0])

            corrected_circuit.compose(source_reflection.inverse(), inplace=True)
            corrected_circuit.rz(np.pi, [0])
            corrected_circuit.compose(circuit, [0, 1], inplace=True)
            corrected_circuit.rz(-np.pi, [0])
            corrected_circuit.compose(source_shift.inverse(), inplace=True)
            corrected_circuit.compose(source_reflection, inplace=True)
            corrected_circuit.global_phase += np.pi / 2

            circuit = corrected_circuit

        circ = QuantumCircuit(2, global_phase=weyl_decomposition.global_phase)

        circ.append(UnitaryGate(weyl_decomposition.K2r), [0])
        circ.append(UnitaryGate(weyl_decomposition.K2l), [1])
        circ.compose(circuit, [0, 1], inplace=True)
        circ.append(UnitaryGate(weyl_decomposition.K1r), [0])
        circ.append(UnitaryGate(weyl_decomposition.K1l), [1])

        circ = self._decomposer1q(circ)

        return circ



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for Fidelity."""

import unittest

import numpy as np

from qiskit.circuit import QuantumCircuit, ParameterVector
from qiskit.circuit.library import RealAmplitudes
from qiskit.primitives import Sampler
from qiskit.algorithms.state_fidelities import ComputeUncompute
from qiskit.test import QiskitTestCase


class TestComputeUncompute(QiskitTestCase):
    """Test Compute-Uncompute Fidelity class"""

    def setUp(self):
        super().setUp()
        parameters = ParameterVector("x", 2)

        rx_rotations = QuantumCircuit(2)
        rx_rotations.rx(parameters[0], 0)
        rx_rotations.rx(parameters[1], 1)

        ry_rotations = QuantumCircuit(2)
        ry_rotations.ry(parameters[0], 0)
        ry_rotations.ry(parameters[1], 1)

        plus = QuantumCircuit(2)
        plus.h([0, 1])

        zero = QuantumCircuit(2)

        rx_rotation = QuantumCircuit(2)
        rx_rotation.rx(parameters[0], 0)
        rx_rotation.h(1)

        self._circuit = [rx_rotations, ry_rotations, plus, zero, rx_rotation]
        self._sampler = Sampler()
        self._left_params = np.array([[0, 0], [np.pi / 2, 0], [0, np.pi / 2], [np.pi, np.pi]])
        self._right_params = np.array([[0, 0], [0, 0], [np.pi / 2, 0], [0, 0]])

    def test_1param_pair(self):
        """test for fidelity with one pair of parameters"""
        fidelity = ComputeUncompute(self._sampler)
        job = fidelity.run(
            self._circuit[0], self._circuit[1], self._left_params[0], self._right_params[0]
        )
        result = job.result()
        np.testing.assert_allclose(result.fidelities, np.array([1.0]))

    def test_4param_pairs(self):
        """test for fidelity with four pairs of parameters"""
        fidelity = ComputeUncompute(self._sampler)
        n = len(self._left_params)
        job = fidelity.run(
            [self._circuit[0]] * n, [self._circuit[1]] * n, self._left_params, self._right_params
        )
        results = job.result()
        np.testing.assert_allclose(results.fidelities, np.array([1.0, 0.5, 0.25, 0.0]), atol=1e-16)

    def test_symmetry(self):
        """test for fidelity with the same circuit"""
        fidelity = ComputeUncompute(self._sampler)
        n = len(self._left_params)
        job_1 = fidelity.run(
            [self._circuit[0]] * n, [self._circuit[0]] * n, self._left_params, self._right_params
        )
        job_2 = fidelity.run(
            [self._circuit[0]] * n, [self._circuit[0]] * n, self._right_params, self._left_params
        )
        results_1 = job_1.result()
        results_2 = job_2.result()
        np.testing.assert_allclose(results_1.fidelities, results_2.fidelities, atol=1e-16)

    def test_no_params(self):
        """test for fidelity without parameters"""
        fidelity = ComputeUncompute(self._sampler)
        job = fidelity.run([self._circuit[2]], [self._circuit[3]])
        results = job.result()
        np.testing.assert_allclose(results.fidelities, np.array([0.25]), atol=1e-16)

        job = fidelity.run([self._circuit[2]], [self._circuit[3]], [], [])
        results = job.result()
        np.testing.assert_allclose(results.fidelities, np.array([0.25]), atol=1e-16)

    def test_left_param(self):
        """test for fidelity with only left parameters"""
        fidelity = ComputeUncompute(self._sampler)
        n = len(self._left_params)
        job = fidelity.run(
            [self._circuit[1]] * n, [self._circuit[3]] * n, values_1=self._left_params
        )
        results = job.result()
        np.testing.assert_allclose(results.fidelities, np.array([1.0, 0.5, 0.5, 0.0]), atol=1e-16)

    def test_right_param(self):
        """test for fidelity with only right parameters"""
        fidelity = ComputeUncompute(self._sampler)
        n = len(self._left_params)
        job = fidelity.run(
            [self._circuit[3]] * n, [self._circuit[1]] * n, values_2=self._left_params
        )
        results = job.result()
        np.testing.assert_allclose(results.fidelities, np.array([1.0, 0.5, 0.5, 0.0]), atol=1e-16)

    def test_not_set_circuits(self):
        """test for fidelity with no circuits."""
        fidelity = ComputeUncompute(self._sampler)
        with self.assertRaises(TypeError):
            job = fidelity.run(
                circuits_1=None,
                circuits_2=None,
                values_1=self._left_params,
                values_2=self._right_params,
            )
            job.result()

    def test_circuit_mismatch(self):
        """test for fidelity with different number of left/right circuits."""
        fidelity = ComputeUncompute(self._sampler)
        n = len(self._left_params)
        with self.assertRaises(ValueError):
            job = fidelity.run(
                [self._circuit[0]] * n,
                [self._circuit[1]] * (n + 1),
                self._left_params,
                self._right_params,
            )
            job.result()

    def test_asymmetric_params(self):
        """test for fidelity when the 2 circuits have different number of
        left/right parameters."""

        fidelity = ComputeUncompute(self._sampler)
        n = len(self._left_params)
        right_params = [[p] for p in self._right_params[:, 0]]
        job = fidelity.run(
            [self._circuit[0]] * n,
            [self._circuit[4]] * n,
            self._left_params,
            right_params,
        )
        result = job.result()
        np.testing.assert_allclose(result.fidelities, np.array([0.5, 0.25, 0.25, 0.0]), atol=1e-16)

    def test_input_format(self):
        """test for different input format variations"""

        fidelity = ComputeUncompute(self._sampler)
        circuit = RealAmplitudes(2)
        values = np.random.random(circuit.num_parameters)
        shift = np.ones_like(values) * 0.01

        # lists of circuits, lists of numpy arrays
        job = fidelity.run([circuit], [circuit], [values], [values + shift])
        result_1 = job.result()

        # lists of circuits, lists of lists
        shift_val = values + shift
        job = fidelity.run([circuit], [circuit], [values.tolist()], [shift_val.tolist()])
        result_2 = job.result()

        # circuits, lists
        shift_val = values + shift
        job = fidelity.run(circuit, circuit, values.tolist(), shift_val.tolist())
        result_3 = job.result()

        # circuits, np.arrays
        job = fidelity.run(circuit, circuit, values, values + shift)
        result_4 = job.result()

        np.testing.assert_allclose(result_1.fidelities, result_2.fidelities, atol=1e-16)
        np.testing.assert_allclose(result_1.fidelities, result_3.fidelities, atol=1e-16)
        np.testing.assert_allclose(result_1.fidelities, result_4.fidelities, atol=1e-16)

    def test_input_measurements(self):
        """test for fidelity with measurements on input circuits"""
        fidelity = ComputeUncompute(self._sampler)
        circuit_1 = self._circuit[0]
        circuit_1.measure_all()
        circuit_2 = self._circuit[1]
        circuit_2.measure_all()

        job = fidelity.run(circuit_1, circuit_2, self._left_params[0], self._right_params[0])
        result = job.result()
        np.testing.assert_allclose(result.fidelities, np.array([1.0]))

    def test_options(self):
        """Test fidelity's run options"""
        sampler_shots = Sampler(options={"shots": 1024})

        with self.subTest("sampler"):
            # Only options in sampler
            fidelity = ComputeUncompute(sampler_shots)
            options = fidelity.options
            job = fidelity.run(self._circuit[2], self._circuit[3])
            result = job.result()
            self.assertEqual(options.__dict__, {"shots": 1024})
            self.assertEqual(result.options.__dict__, {"shots": 1024})

        with self.subTest("fidelity init"):
            # Fidelity default options override sampler
            # options and add new fields
            fidelity = ComputeUncompute(sampler_shots, options={"shots": 2048, "dummy": 100})
            options = fidelity.options
            job = fidelity.run(self._circuit[2], self._circuit[3])
            result = job.result()
            self.assertEqual(options.__dict__, {"shots": 2048, "dummy": 100})
            self.assertEqual(result.options.__dict__, {"shots": 2048, "dummy": 100})

        with self.subTest("fidelity update"):
            # Update fidelity options
            fidelity = ComputeUncompute(sampler_shots, options={"shots": 2048, "dummy": 100})
            fidelity.update_default_options(shots=100)
            options = fidelity.options
            job = fidelity.run(self._circuit[2], self._circuit[3])
            result = job.result()
            self.assertEqual(options.__dict__, {"shots": 100, "dummy": 100})
            self.assertEqual(result.options.__dict__, {"shots": 100, "dummy": 100})

        with self.subTest("fidelity run"):
            # Run options override fidelity options
            fidelity = ComputeUncompute(sampler_shots, options={"shots": 2048, "dummy": 100})
            job = fidelity.run(self._circuit[2], self._circuit[3], shots=50, dummy=None)
            options = fidelity.options
            result = job.result()
            # Only default + sampler options. Not run.
            self.assertEqual(options.__dict__, {"shots": 2048, "dummy": 100})
            self.assertEqual(result.options.__dict__, {"shots": 50, "dummy": None})


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Assembler Test."""

import unittest
import io
from logging import StreamHandler, getLogger
import sys
import copy

import numpy as np
from qiskit import pulse
from qiskit.circuit import Instruction, Gate, Parameter, ParameterVector
from qiskit.circuit import QuantumRegister, ClassicalRegister, QuantumCircuit
from qiskit.compiler.assembler import assemble
from qiskit.exceptions import QiskitError
from qiskit.pulse import Schedule, Acquire, Play
from qiskit.pulse.channels import MemorySlot, AcquireChannel, DriveChannel, MeasureChannel
from qiskit.pulse.configuration import Kernel, Discriminator
from qiskit.pulse.library import gaussian
from qiskit.qobj import QasmQobj, PulseQobj
from qiskit.qobj.utils import MeasLevel, MeasReturnType
from qiskit.pulse.macros import measure
from qiskit.test import QiskitTestCase
from qiskit.providers.fake_provider import (
    FakeOpenPulse2Q,
    FakeOpenPulse3Q,
    FakeYorktown,
    FakeHanoi,
)


class RxGate(Gate):
    """Used to test custom gate assembly.

    Useful for testing pulse gates with parameters, as well.
    Note: Parallel maps (e.g., in assemble_circuits) pickle their input,
          so circuit features have to be defined top level.
    """

    def __init__(self, theta):
        super().__init__("rxtheta", 1, [theta])


class TestCircuitAssembler(QiskitTestCase):
    """Tests for assembling circuits to qobj."""

    def setUp(self):
        super().setUp()
        qr = QuantumRegister(2, name="q")
        cr = ClassicalRegister(2, name="c")
        self.circ = QuantumCircuit(qr, cr, name="circ")
        self.circ.h(qr[0])
        self.circ.cx(qr[0], qr[1])
        self.circ.measure(qr, cr)

        self.backend = FakeYorktown()
        self.backend_config = self.backend.configuration()
        self.num_qubits = self.backend_config.n_qubits

        # lo test values
        self.default_qubit_lo_freq = [5e9 for _ in range(self.num_qubits)]
        self.default_meas_lo_freq = [6.7e9 for _ in range(self.num_qubits)]
        self.user_lo_config_dict = {
            pulse.DriveChannel(0): 5.55e9,
            pulse.MeasureChannel(0): 6.64e9,
            pulse.DriveChannel(3): 4.91e9,
            pulse.MeasureChannel(4): 6.1e9,
        }
        self.user_lo_config = pulse.LoConfig(self.user_lo_config_dict)

    def test_assemble_single_circuit(self):
        """Test assembling a single circuit."""
        qobj = assemble(self.circ, shots=2000, memory=True)

        self.assertIsInstance(qobj, QasmQobj)
        self.assertEqual(qobj.config.shots, 2000)
        self.assertEqual(qobj.config.memory, True)
        self.assertEqual(len(qobj.experiments), 1)
        self.assertEqual(qobj.experiments[0].instructions[1].name, "cx")

    def test_assemble_multiple_circuits(self):
        """Test assembling multiple circuits, all should have the same config."""
        qr0 = QuantumRegister(2, name="q0")
        qc0 = ClassicalRegister(2, name="c0")
        circ0 = QuantumCircuit(qr0, qc0, name="circ0")
        circ0.h(qr0[0])
        circ0.cx(qr0[0], qr0[1])
        circ0.measure(qr0, qc0)

        qr1 = QuantumRegister(3, name="q1")
        qc1 = ClassicalRegister(3, name="c1")
        circ1 = QuantumCircuit(qr1, qc1, name="circ0")
        circ1.h(qr1[0])
        circ1.cx(qr1[0], qr1[1])
        circ1.cx(qr1[0], qr1[2])
        circ1.measure(qr1, qc1)

        qobj = assemble([circ0, circ1], shots=100, memory=False, seed_simulator=6)

        self.assertIsInstance(qobj, QasmQobj)
        self.assertEqual(qobj.config.seed_simulator, 6)
        self.assertEqual(len(qobj.experiments), 2)
        self.assertEqual(qobj.experiments[1].config.n_qubits, 3)
        self.assertEqual(len(qobj.experiments), 2)
        self.assertEqual(len(qobj.experiments[1].instructions), 6)

    def test_assemble_no_run_config(self):
        """Test assembling with no run_config, relying on default."""
        qobj = assemble(self.circ)

        self.assertIsInstance(qobj, QasmQobj)
        self.assertEqual(qobj.config.shots, 1024)

    def test_shots_greater_than_max_shots(self):
        """Test assembling with shots greater than max shots"""
        self.assertRaises(QiskitError, assemble, self.backend, shots=1024000)

    def test_shots_not_of_type_int(self):
        """Test assembling with shots having type other than int"""
        self.assertRaises(QiskitError, assemble, self.backend, shots="1024")

    def test_shots_of_type_numpy_int64(self):
        """Test assembling with shots having type numpy.int64"""
        qobj = assemble(self.circ, shots=np.int64(2048))
        self.assertEqual(qobj.config.shots, 2048)

    def test_default_shots_greater_than_max_shots(self):
        """Test assembling with default shots greater than max shots"""
        self.backend_config.max_shots = 5

        qobj = assemble(self.circ, self.backend)

        self.assertIsInstance(qobj, QasmQobj)
        self.assertEqual(qobj.config.shots, 5)

    def test_assemble_initialize(self):
        """Test assembling a circuit with an initialize."""
        q = QuantumRegister(2, name="q")
        circ = QuantumCircuit(q, name="circ")
        circ.initialize([1 / np.sqrt(2), 0, 0, 1 / np.sqrt(2)], q[:])

        qobj = assemble(circ)

        self.assertIsInstance(qobj, QasmQobj)
        self.assertEqual(qobj.experiments[0].instructions[0].name, "initialize")
        np.testing.assert_almost_equal(
            qobj.experiments[0].instructions[0].params, [0.7071067811865, 0, 0, 0.707106781186]
        )

    def test_assemble_meas_level_meas_return(self):
        """Test assembling a circuit schedule with `meas_level`."""
        qobj = assemble(self.circ, meas_level=1, meas_return="single")

        self.assertIsInstance(qobj, QasmQobj)
        self.assertEqual(qobj.config.meas_level, 1)
        self.assertEqual(qobj.config.meas_return, "single")

        # no meas_level set
        qobj = assemble(self.circ)

        self.assertIsInstance(qobj, QasmQobj)
        self.assertEqual(qobj.config.meas_level, 2)
        self.assertEqual(hasattr(qobj.config, "meas_return"), False)

    def test_assemble_backend_rep_delays(self):
        """Check that rep_delay is properly set from backend values."""
        rep_delay_range = [2.5e-3, 4.5e-3]  # sec
        default_rep_delay = 3.0e-3
        setattr(self.backend_config, "rep_delay_range", rep_delay_range)
        setattr(self.backend_config, "default_rep_delay", default_rep_delay)

        # dynamic rep rates off
        setattr(self.backend_config, "dynamic_reprate_enabled", False)
        qobj = assemble(self.circ, self.backend)
        self.assertEqual(hasattr(qobj.config, "rep_delay"), False)

        # dynamic rep rates on
        setattr(self.backend_config, "dynamic_reprate_enabled", True)
        qobj = assemble(self.circ, self.backend)
        self.assertEqual(qobj.config.rep_delay, default_rep_delay * 1e6)

    def test_assemble_user_rep_time_delay(self):
        """Check that user runtime config rep_delay works."""
        # set custom rep_delay in runtime config
        rep_delay = 2.2e-6
        rep_delay_range = [0, 3e-6]  # sec
        setattr(self.backend_config, "rep_delay_range", rep_delay_range)

        # dynamic rep rates off (no default so shouldn't be in qobj config)
        setattr(self.backend_config, "dynamic_reprate_enabled", False)
        qobj = assemble(self.circ, self.backend, rep_delay=rep_delay)
        self.assertEqual(hasattr(qobj.config, "rep_delay"), False)

        # turn on dynamic rep rates, rep_delay should be set
        setattr(self.backend_config, "dynamic_reprate_enabled", True)
        qobj = assemble(self.circ, self.backend, rep_delay=rep_delay)
        self.assertEqual(qobj.config.rep_delay, 2.2)

        # test ``rep_delay=0``
        qobj = assemble(self.circ, self.backend, rep_delay=0)
        self.assertEqual(qobj.config.rep_delay, 0)

        # use ``rep_delay`` outside of ``rep_delay_range```
        rep_delay_large = 5.0e-6
        with self.assertRaises(QiskitError):
            assemble(self.circ, self.backend, rep_delay=rep_delay_large)

    def test_assemble_opaque_inst(self):
        """Test opaque instruction is assembled as-is"""
        opaque_inst = Instruction(name="my_inst", num_qubits=4, num_clbits=2, params=[0.5, 0.4])
        q = QuantumRegister(6, name="q")
        c = ClassicalRegister(4, name="c")
        circ = QuantumCircuit(q, c, name="circ")
        circ.append(opaque_inst, [q[0], q[2], q[5], q[3]], [c[3], c[0]])

        qobj = assemble(circ)

        self.assertIsInstance(qobj, QasmQobj)
        self.assertEqual(len(qobj.experiments[0].instructions), 1)
        self.assertEqual(qobj.experiments[0].instructions[0].name, "my_inst")
        self.assertEqual(qobj.experiments[0].instructions[0].qubits, [0, 2, 5, 3])
        self.assertEqual(qobj.experiments[0].instructions[0].memory, [3, 0])
        self.assertEqual(qobj.experiments[0].instructions[0].params, [0.5, 0.4])

    def test_assemble_unroll_parametervector(self):
        """Verfiy that assemble unrolls parametervectors ref #5467"""
        pv1 = ParameterVector("pv1", 3)
        pv2 = ParameterVector("pv2", 3)
        qc = QuantumCircuit(2, 2)
        for i in range(3):
            qc.rx(pv1[i], 0)
            qc.ry(pv2[i], 1)
        qc.barrier()
        qc.measure([0, 1], [0, 1])

        qc.bind_parameters({pv1: [0.1, 0.2, 0.3], pv2: [0.4, 0.5, 0.6]})

        qobj = assemble(qc, parameter_binds=[{pv1: [0.1, 0.2, 0.3], pv2: [0.4, 0.5, 0.6]}])

        self.assertIsInstance(qobj, QasmQobj)
        self.assertEqual(qobj.experiments[0].instructions[0].params[0], 0.100000000000000)
        self.assertEqual(qobj.experiments[0].instructions[1].params[0], 0.400000000000000)
        self.assertEqual(qobj.experiments[0].instructions[2].params[0], 0.200000000000000)
        self.assertEqual(qobj.experiments[0].instructions[3].params[0], 0.500000000000000)
        self.assertEqual(qobj.experiments[0].instructions[4].params[0], 0.300000000000000)
        self.assertEqual(qobj.experiments[0].instructions[5].params[0], 0.600000000000000)

    def test_measure_to_registers_when_conditionals(self):
        """Verify assemble_circuits maps all measure ops on to a register slot
        for a circuit containing conditionals."""
        qr = QuantumRegister(2)
        cr1 = ClassicalRegister(1)
        cr2 = ClassicalRegister(2)
        qc = QuantumCircuit(qr, cr1, cr2)

        qc.measure(qr[0], cr1)  # Measure not required for a later conditional
        qc.measure(qr[1], cr2[1])  # Measure required for a later conditional
        qc.h(qr[1]).c_if(cr2, 3)

        qobj = assemble(qc)

        first_measure, second_measure = (
            op for op in qobj.experiments[0].instructions if op.name == "measure"
        )

        self.assertTrue(hasattr(first_measure, "register"))
        self.assertEqual(first_measure.register, first_measure.memory)
        self.assertTrue(hasattr(second_measure, "register"))
        self.assertEqual(second_measure.register, second_measure.memory)

    def test_convert_to_bfunc_plus_conditional(self):
        """Verify assemble_circuits converts conditionals from QASM to Qobj."""
        qr = QuantumRegister(1)
        cr = ClassicalRegister(1)
        qc = QuantumCircuit(qr, cr)

        qc.h(qr[0]).c_if(cr, 1)

        qobj = assemble(qc)

        bfunc_op, h_op = qobj.experiments[0].instructions

        self.assertEqual(bfunc_op.name, "bfunc")
        self.assertEqual(bfunc_op.mask, "0x1")
        self.assertEqual(bfunc_op.val, "0x1")
        self.assertEqual(bfunc_op.relation, "==")

        self.assertTrue(hasattr(h_op, "conditional"))
        self.assertEqual(bfunc_op.register, h_op.conditional)

    def test_convert_to_bfunc_plus_conditional_onebit(self):
        """Verify assemble_circuits converts single bit conditionals from QASM to Qobj."""
        qr = QuantumRegister(1)
        cr = ClassicalRegister(3)
        qc = QuantumCircuit(qr, cr)

        qc.h(qr[0]).c_if(cr[2], 1)

        qobj = assemble(qc)

        inst_set = qobj.experiments[0].instructions
        [bfunc_op, h_op] = inst_set

        self.assertEqual(len(inst_set), 2)
        self.assertEqual(bfunc_op.name, "bfunc")
        self.assertEqual(bfunc_op.mask, "0x4")
        self.assertEqual(bfunc_op.val, "0x4")
        self.assertEqual(bfunc_op.relation, "==")

        self.assertTrue(hasattr(h_op, "conditional"))
        self.assertEqual(bfunc_op.register, h_op.conditional)

    def test_resize_value_to_register(self):
        """Verify assemble_circuits converts the value provided on the classical
        creg to its mapped location on the device register."""
        qr = QuantumRegister(1)
        cr1 = ClassicalRegister(2)
        cr2 = ClassicalRegister(2)
        cr3 = ClassicalRegister(1)
        qc = QuantumCircuit(qr, cr1, cr2, cr3)

        qc.h(qr[0]).c_if(cr2, 2)

        qobj = assemble(qc)

        bfunc_op, h_op = qobj.experiments[0].instructions

        self.assertEqual(bfunc_op.name, "bfunc")
        self.assertEqual(bfunc_op.mask, "0xC")
        self.assertEqual(bfunc_op.val, "0x8")
        self.assertEqual(bfunc_op.relation, "==")

        self.assertTrue(hasattr(h_op, "conditional"))
        self.assertEqual(bfunc_op.register, h_op.conditional)

    def test_assemble_circuits_raises_for_bind_circuit_mismatch(self):
        """Verify assemble_circuits raises error for parameterized circuits without matching
        binds."""
        qr = QuantumRegister(2)
        x = Parameter("x")
        y = Parameter("y")

        full_bound_circ = QuantumCircuit(qr)
        full_param_circ = QuantumCircuit(qr)
        partial_param_circ = QuantumCircuit(qr)

        partial_param_circ.p(x, qr[0])

        full_param_circ.p(x, qr[0])
        full_param_circ.p(y, qr[1])

        partial_bind_args = {"parameter_binds": [{x: 1}, {x: 0}]}
        full_bind_args = {"parameter_binds": [{x: 1, y: 1}, {x: 0, y: 0}]}
        inconsistent_bind_args = {"parameter_binds": [{x: 1}, {x: 0, y: 0}]}

        # Raise when parameters passed for non-parametric circuit
        self.assertRaises(QiskitError, assemble, full_bound_circ, **partial_bind_args)

        # Raise when no parameters passed for parametric circuit
        self.assertRaises(QiskitError, assemble, partial_param_circ)
        self.assertRaises(QiskitError, assemble, full_param_circ)

        # Raise when circuit has more parameters than run_config
        self.assertRaises(QiskitError, assemble, full_param_circ, **partial_bind_args)

        # Raise when not all circuits have all parameters
        self.assertRaises(
            QiskitError, assemble, [full_param_circ, partial_param_circ], **full_bind_args
        )

        # Raise when not all binds have all circuit params
        self.assertRaises(QiskitError, assemble, full_param_circ, **inconsistent_bind_args)

    def test_assemble_circuits_rases_for_bind_mismatch_over_expressions(self):
        """Verify assemble_circuits raises for invalid binds for circuit including
        ParameterExpressions.
        """
        qr = QuantumRegister(1)
        x = Parameter("x")
        y = Parameter("y")

        expr_circ = QuantumCircuit(qr)

        expr_circ.p(x + y, qr[0])

        partial_bind_args = {"parameter_binds": [{x: 1}, {x: 0}]}

        # Raise when no parameters passed for parametric circuit
        self.assertRaises(QiskitError, assemble, expr_circ)

        # Raise when circuit has more parameters than run_config
        self.assertRaises(QiskitError, assemble, expr_circ, **partial_bind_args)

    def test_assemble_circuits_binds_parameters(self):
        """Verify assemble_circuits applies parameter bindings and output circuits are bound."""
        qr = QuantumRegister(1)
        qc1 = QuantumCircuit(qr)
        qc2 = QuantumCircuit(qr)
        qc3 = QuantumCircuit(qr)

        x = Parameter("x")
        y = Parameter("y")
        sum_ = x + y
        product_ = x * y

        qc1.u(x, y, 0, qr[0])

        qc2.rz(x, qr[0])
        qc2.rz(y, qr[0])

        qc3.u(sum_, product_, 0, qr[0])

        bind_args = {"parameter_binds": [{x: 0, y: 0}, {x: 1, y: 0}, {x: 1, y: 1}]}

        qobj = assemble([qc1, qc2, qc3], **bind_args)

        self.assertEqual(len(qobj.experiments), 9)
        self.assertEqual(
            [len(expt.instructions) for expt in qobj.experiments], [1, 1, 1, 2, 2, 2, 1, 1, 1]
        )

        def _qobj_inst_params(expt_no, inst_no):
            expt = qobj.experiments[expt_no]
            inst = expt.instructions[inst_no]
            return [float(p) for p in inst.params]

        self.assertEqual(_qobj_inst_params(0, 0), [0, 0, 0])
        self.assertEqual(_qobj_inst_params(1, 0), [1, 0, 0])
        self.assertEqual(_qobj_inst_params(2, 0), [1, 1, 0])

        self.assertEqual(_qobj_inst_params(3, 0), [0])
        self.assertEqual(_qobj_inst_params(3, 1), [0])
        self.assertEqual(_qobj_inst_params(4, 0), [1])
        self.assertEqual(_qobj_inst_params(4, 1), [0])
        self.assertEqual(_qobj_inst_params(5, 0), [1])
        self.assertEqual(_qobj_inst_params(5, 1), [1])

        self.assertEqual(_qobj_inst_params(6, 0), [0, 0, 0])
        self.assertEqual(_qobj_inst_params(7, 0), [1, 0, 0])
        self.assertEqual(_qobj_inst_params(8, 0), [2, 1, 0])

    def test_init_qubits_default(self):
        """Check that the init_qubits=None assemble option is passed on to the qobj."""
        qobj = assemble(self.circ)
        self.assertEqual(qobj.config.init_qubits, True)

    def test_init_qubits_true(self):
        """Check that the init_qubits=True assemble option is passed on to the qobj."""
        qobj = assemble(self.circ, init_qubits=True)
        self.assertEqual(qobj.config.init_qubits, True)

    def test_init_qubits_false(self):
        """Check that the init_qubits=False assemble option is passed on to the qobj."""
        qobj = assemble(self.circ, init_qubits=False)
        self.assertEqual(qobj.config.init_qubits, False)

    def test_circuit_with_global_phase(self):
        """Test that global phase for a circuit is handled correctly."""
        circ = QuantumCircuit(2)
        circ.h(0)
        circ.cx(0, 1)
        circ.measure_all()
        circ.global_phase = 0.3 * np.pi
        qobj = assemble([circ, self.circ])
        self.assertEqual(getattr(qobj.experiments[1].header, "global_phase"), 0)
        self.assertEqual(getattr(qobj.experiments[0].header, "global_phase"), 0.3 * np.pi)

    def test_circuit_global_phase_gate_definitions(self):
        """Test circuit with global phase on gate definitions."""

        class TestGate(Gate):
            """dummy gate"""

            def __init__(self):
                super().__init__("test_gate", 1, [])

            def _define(self):
                circ_def = QuantumCircuit(1)
                circ_def.x(0)
                circ_def.global_phase = np.pi
                self._definition = circ_def

        gate = TestGate()
        circ = QuantumCircuit(1)
        circ.append(gate, [0])
        qobj = assemble([circ])
        self.assertEqual(getattr(qobj.experiments[0].header, "global_phase"), 0)
        circ.global_phase = np.pi / 2
        qobj = assemble([circ])
        self.assertEqual(getattr(qobj.experiments[0].header, "global_phase"), np.pi / 2)

    def test_pulse_gates_single_circ(self):
        """Test that we can add calibrations to circuits."""
        theta = Parameter("theta")
        circ = QuantumCircuit(2)
        circ.h(0)
        circ.append(RxGate(3.14), [0])
        circ.append(RxGate(theta), [1])
        circ = circ.assign_parameters({theta: 3.14})

        with pulse.build() as custom_h_schedule:
            pulse.play(pulse.library.Drag(50, 0.15, 4, 2), pulse.DriveChannel(0))

        with pulse.build() as x180:
            pulse.play(pulse.library.Gaussian(50, 0.2, 5), pulse.DriveChannel(1))

        circ.add_calibration("h", [0], custom_h_schedule)
        circ.add_calibration(RxGate(3.14), [0], x180)
        circ.add_calibration(RxGate(3.14), [1], x180)

        qobj = assemble(circ, FakeOpenPulse2Q())
        # Only one circuit, so everything is stored at the job level
        cals = qobj.config.calibrations
        lib = qobj.config.pulse_library
        self.assertFalse(hasattr(qobj.experiments[0].config, "calibrations"))
        self.assertEqual([gate.name == "rxtheta" for gate in cals.gates].count(True), 2)
        self.assertEqual([gate.name == "h" for gate in cals.gates].count(True), 1)
        self.assertEqual(len(lib), 2)
        self.assertTrue(all(len(item.samples) == 50 for item in lib))

    def test_pulse_gates_with_parameteric_pulses(self):
        """Test that pulse gates are assembled efficiently for backends that enable
        parametric pulses.
        """
        with pulse.build() as custom_h_schedule:
            pulse.play(pulse.library.Drag(50, 0.15, 4, 2), pulse.DriveChannel(0))

        circ = QuantumCircuit(2)
        circ.h(0)
        circ.add_calibration("h", [0], custom_h_schedule)

        backend = FakeOpenPulse2Q()
        backend.configuration().parametric_pulses = ["drag"]
        qobj = assemble(circ, backend)
        self.assertFalse(hasattr(qobj.config, "pulse_library"))
        self.assertTrue(hasattr(qobj.config, "calibrations"))

    def test_pulse_gates_multiple_circuits(self):
        """Test one circuit with cals and another without."""
        with pulse.build() as dummy_sched:
            pulse.play(pulse.library.Drag(50, 0.15, 4, 2), pulse.DriveChannel(0))

        circ = QuantumCircuit(2)
        circ.h(0)
        circ.append(RxGate(3.14), [1])
        circ.add_calibration("h", [0], dummy_sched)
        circ.add_calibration(RxGate(3.14), [1], dummy_sched)

        circ2 = QuantumCircuit(2)
        circ2.h(0)

        qobj = assemble([circ, circ2], FakeOpenPulse2Q())
        self.assertEqual(len(qobj.config.pulse_library), 1)
        self.assertEqual(len(qobj.experiments[0].config.calibrations.gates), 2)
        self.assertFalse(hasattr(qobj.config, "calibrations"))
        self.assertFalse(hasattr(qobj.experiments[1].config, "calibrations"))

    def test_pulse_gates_common_cals(self):
        """Test that common calibrations are added at the top level."""
        with pulse.build() as dummy_sched:
            pulse.play(pulse.library.Drag(50, 0.15, 4, 2), pulse.DriveChannel(0))

        circ = QuantumCircuit(2)
        circ.h(0)
        circ.append(RxGate(3.14), [1])
        circ.add_calibration("h", [0], dummy_sched)
        circ.add_calibration(RxGate(3.14), [1], dummy_sched)

        circ2 = QuantumCircuit(2)
        circ2.h(0)
        circ2.add_calibration(RxGate(3.14), [1], dummy_sched)

        qobj = assemble([circ, circ2], FakeOpenPulse2Q())
        # Identical pulses are only added once
        self.assertEqual(len(qobj.config.pulse_library), 1)
        # Identical calibrations are only added once
        self.assertEqual(qobj.config.calibrations.gates[0].name, "rxtheta")
        self.assertEqual(qobj.config.calibrations.gates[0].params, [3.14])
        self.assertEqual(qobj.config.calibrations.gates[0].qubits, [1])
        self.assertEqual(len(qobj.experiments[0].config.calibrations.gates), 1)
        self.assertFalse(hasattr(qobj.experiments[1].config, "calibrations"))

    def test_assemble_adds_circuit_metadata_to_experiment_header(self):
        """Verify that any circuit metadata is added to the exeriment header."""
        circ = QuantumCircuit(2, metadata=dict(experiment_type="gst", execution_number="1234"))
        qobj = assemble(circ, shots=100, memory=False, seed_simulator=6)
        self.assertEqual(
            qobj.experiments[0].header.metadata,
            {"experiment_type": "gst", "execution_number": "1234"},
        )

    def test_pulse_gates_delay_only(self):
        """Test that a single delay gate is translated to an instruction."""
        circ = QuantumCircuit(2)
        circ.append(Gate("test", 1, []), [0])
        test_sched = pulse.Delay(64, DriveChannel(0)) + pulse.Delay(160, DriveChannel(0))
        circ.add_calibration("test", [0], test_sched)
        qobj = assemble(circ, FakeOpenPulse2Q())
        self.assertEqual(len(qobj.config.calibrations.gates[0].instructions), 2)
        self.assertEqual(
            qobj.config.calibrations.gates[0].instructions[1].to_dict(),
            {"name": "delay", "t0": 64, "ch": "d0", "duration": 160},
        )

    def test_job_qubit_meas_los_no_range(self):
        """Test that adding job qubit/meas lo freq lists are assembled into the qobj.config, w/ out
        any lo range."""
        qobj = assemble(
            self.circ,
            backend=self.backend,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
        )

        # convert to ghz
        qubit_lo_freq_ghz = [freq / 1e9 for freq in self.default_qubit_lo_freq]
        meas_lo_freq_ghz = [freq / 1e9 for freq in self.default_meas_lo_freq]
        self.assertEqual(qobj.config.qubit_lo_freq, qubit_lo_freq_ghz)
        self.assertEqual(qobj.config.meas_lo_freq, meas_lo_freq_ghz)

    def test_job_lo_errors(self):
        """Test that job lo's are checked against the lo ranges and that errors are thrown if either
        quantity has an incorrect length or type."""
        qubit_lo_range = [[freq - 5e6, freq + 5e6] for freq in self.default_qubit_lo_freq]
        meas_lo_range = [[freq - 5e6, freq + 5e6] for freq in self.default_meas_lo_freq]

        # lo range not a nested list
        with self.assertRaises(QiskitError):
            assemble(
                self.circ,
                backend=self.backend,
                qubit_lo_freq=self.default_qubit_lo_freq,
                meas_lo_freq=self.default_meas_lo_freq,
                qubit_lo_range=[4.995e9 for i in range(self.num_qubits)],
                meas_lo_range=meas_lo_range,
            )

        # qubit lo range inner list not 2d
        with self.assertRaises(QiskitError):
            assemble(
                self.circ,
                backend=self.backend,
                qubit_lo_freq=self.default_qubit_lo_freq,
                meas_lo_freq=self.default_meas_lo_freq,
                qubit_lo_range=qubit_lo_range,
                meas_lo_range=[[6.695e9] for i in range(self.num_qubits)],
            )

        # meas lo range inner list not 2d
        with self.assertRaises(QiskitError):
            assemble(
                self.circ,
                backend=self.backend,
                qubit_lo_freq=self.default_qubit_lo_freq,
                meas_lo_freq=self.default_meas_lo_freq,
                qubit_lo_range=qubit_lo_range,
                meas_lo_range=[[6.695e9] for i in range(self.num_qubits)],
            )

        # qubit lo out of range
        with self.assertRaises(QiskitError):
            assemble(
                self.circ,
                backend=self.backend,
                qubit_lo_freq=self.default_qubit_lo_freq,
                meas_lo_freq=self.default_meas_lo_freq,
                qubit_lo_range=[[5.005e9, 5.010e9] for i in range(self.num_qubits)],
                meas_lo_range=meas_lo_range,
            )

        # meas lo out of range
        with self.assertRaises(QiskitError):
            assemble(
                self.circ,
                backend=self.backend,
                qubit_lo_freq=self.default_qubit_lo_freq,
                meas_lo_freq=self.default_meas_lo_freq,
                qubit_lo_range=qubit_lo_range,
                meas_lo_range=[[6.705e9, 6.710e9] for i in range(self.num_qubits)],
            )

    def test_job_qubit_meas_los_w_range(self):
        """Test that adding job qubit/meas lo freq lists are assembled into the qobj.config, w/ lo
        ranges input. Verify that lo ranges do not enter into the config."""
        qubit_lo_range = [[freq - 5e6, freq + 5e6] for freq in self.default_qubit_lo_freq]
        meas_lo_range = [[freq - 5e6, freq + 5e6] for freq in self.default_meas_lo_freq]
        qobj = assemble(
            self.circ,
            backend=self.backend,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            qubit_lo_range=qubit_lo_range,
            meas_lo_range=meas_lo_range,
        )

        # convert to ghz
        qubit_lo_freq_ghz = [freq / 1e9 for freq in self.default_qubit_lo_freq]
        meas_lo_freq_ghz = [freq / 1e9 for freq in self.default_meas_lo_freq]
        self.assertEqual(qobj.config.qubit_lo_freq, qubit_lo_freq_ghz)
        self.assertEqual(qobj.config.meas_lo_freq, meas_lo_freq_ghz)
        self.assertNotIn("qubit_lo_range", qobj.config.to_dict())
        self.assertNotIn("meas_lo_range", qobj.config.to_dict())

    def test_assemble_single_circ_single_lo_config(self):
        """Test assembling a single circuit, with a single experiment level lo config."""
        qobj = assemble(
            self.circ,
            self.backend,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=self.user_lo_config,
        )

        self.assertListEqual(qobj.config.qubit_lo_freq, [5.55, 5, 5, 4.91, 5])
        self.assertListEqual(qobj.config.meas_lo_freq, [6.64, 6.7, 6.7, 6.7, 6.1])
        self.assertEqual(len(qobj.experiments), 1)

    def test_assemble_single_circ_single_lo_config_dict(self):
        """Test assembling a single circuit, with a single experiment level lo config supplied as
        dictionary."""
        qobj = assemble(
            self.circ,
            self.backend,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=self.user_lo_config_dict,
        )

        self.assertListEqual(qobj.config.qubit_lo_freq, [5.55, 5, 5, 4.91, 5])
        self.assertListEqual(qobj.config.meas_lo_freq, [6.64, 6.7, 6.7, 6.7, 6.1])
        self.assertEqual(len(qobj.experiments), 1)

    def test_assemble_single_circ_multi_lo_config(self):
        """Test assembling a single circuit, with multiple experiment level lo configs (frequency
        sweep).
        """
        user_lo_config_dict2 = {
            pulse.DriveChannel(1): 5.55e9,
            pulse.MeasureChannel(1): 6.64e9,
            pulse.DriveChannel(4): 4.91e9,
            pulse.MeasureChannel(3): 6.1e9,
        }
        user_lo_config2 = pulse.LoConfig(user_lo_config_dict2)
        qobj = assemble(
            self.circ,
            self.backend,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=[self.user_lo_config, user_lo_config2],
        )

        qubit_lo_freq_ghz = [freq / 1e9 for freq in self.default_qubit_lo_freq]
        meas_lo_freq_ghz = [freq / 1e9 for freq in self.default_meas_lo_freq]
        self.assertListEqual(qobj.config.qubit_lo_freq, qubit_lo_freq_ghz)
        self.assertListEqual(qobj.config.meas_lo_freq, meas_lo_freq_ghz)
        self.assertEqual(len(qobj.experiments), 2)

        # experiment 0 los
        self.assertEqual(qobj.experiments[0].config.qubit_lo_freq, [5.55, 5, 5, 4.91, 5])
        self.assertEqual(qobj.experiments[0].config.meas_lo_freq, [6.64, 6.7, 6.7, 6.7, 6.1])
        # experiment 1 los
        self.assertEqual(qobj.experiments[1].config.qubit_lo_freq, [5, 5.55, 5, 5, 4.91])
        self.assertEqual(qobj.experiments[1].config.meas_lo_freq, [6.7, 6.64, 6.7, 6.1, 6.7])

    def test_assemble_multi_circ_multi_lo_config(self):
        """Test assembling circuits, with the same number of experiment level lo configs (n:n
        setup)."""
        user_lo_config_dict2 = {
            pulse.DriveChannel(1): 5.55e9,
            pulse.MeasureChannel(1): 6.64e9,
            pulse.DriveChannel(4): 4.91e9,
            pulse.MeasureChannel(3): 6.1e9,
        }
        user_lo_config2 = pulse.LoConfig(user_lo_config_dict2)
        qobj = assemble(
            [self.circ, self.circ],
            self.backend,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=[self.user_lo_config, user_lo_config2],
        )

        qubit_lo_freq_ghz = [freq / 1e9 for freq in self.default_qubit_lo_freq]
        meas_lo_freq_ghz = [freq / 1e9 for freq in self.default_meas_lo_freq]
        self.assertListEqual(qobj.config.qubit_lo_freq, qubit_lo_freq_ghz)
        self.assertListEqual(qobj.config.meas_lo_freq, meas_lo_freq_ghz)
        self.assertEqual(len(qobj.experiments), 2)

        # experiment 0 los
        self.assertEqual(qobj.experiments[0].config.qubit_lo_freq, [5.55, 5, 5, 4.91, 5])
        self.assertEqual(qobj.experiments[0].config.meas_lo_freq, [6.64, 6.7, 6.7, 6.7, 6.1])
        # experiment 1 los
        self.assertEqual(qobj.experiments[1].config.qubit_lo_freq, [5, 5.55, 5, 5, 4.91])
        self.assertEqual(qobj.experiments[1].config.meas_lo_freq, [6.7, 6.64, 6.7, 6.1, 6.7])

    def test_assemble_multi_circ_single_lo_config(self):
        """Test assembling multiple circuits, with a single experiment level lo config (should
        override job level)."""
        qobj = assemble(
            [self.circ, self.circ],
            self.backend,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=self.user_lo_config,
        )

        self.assertListEqual(qobj.config.qubit_lo_freq, [5.55, 5, 5, 4.91, 5])
        self.assertListEqual(qobj.config.meas_lo_freq, [6.64, 6.7, 6.7, 6.7, 6.1])
        self.assertEqual(len(qobj.experiments), 2)

    def test_assemble_multi_circ_wrong_number_of_multi_lo_configs(self):
        """Test assembling circuits, with a different number of experiment level lo configs (n:m
        setup).
        """
        with self.assertRaises(QiskitError):
            assemble(
                [self.circ, self.circ, self.circ],
                self.backend,
                qubit_lo_freq=self.default_qubit_lo_freq,
                meas_lo_freq=self.default_meas_lo_freq,
                schedule_los=[self.user_lo_config, self.user_lo_config],
            )

    def test_assemble_circ_lo_config_errors(self):
        """Test that lo config errors are raised properly if experiment level los are provided and
        some are missing or if default values are not provided. Also check that experiment level lo
        range is validated."""
        # no defaults, but have drive/meas experiment level los for each qubit (no error)
        full_lo_config_dict = {
            pulse.DriveChannel(0): 4.85e9,
            pulse.DriveChannel(1): 4.9e9,
            pulse.DriveChannel(2): 4.95e9,
            pulse.DriveChannel(3): 5e9,
            pulse.DriveChannel(4): 5.05e9,
            pulse.MeasureChannel(0): 6.8e9,
            pulse.MeasureChannel(1): 6.85e9,
            pulse.MeasureChannel(2): 6.9e9,
            pulse.MeasureChannel(3): 6.95e9,
            pulse.MeasureChannel(4): 7e9,
        }

        qobj = assemble(self.circ, self.backend, schedule_los=full_lo_config_dict)

        self.assertListEqual(qobj.config.qubit_lo_freq, [4.85, 4.9, 4.95, 5, 5.05])
        self.assertListEqual(qobj.config.meas_lo_freq, [6.8, 6.85, 6.9, 6.95, 7])
        self.assertEqual(len(qobj.experiments), 1)

        # no defaults and missing experiment level drive lo raises
        missing_drive_lo_config_dict = copy.deepcopy(full_lo_config_dict)
        missing_drive_lo_config_dict.pop(pulse.DriveChannel(0))
        with self.assertRaises(QiskitError):
            qobj = assemble(self.circ, self.backend, schedule_los=missing_drive_lo_config_dict)

        # no defaults and missing experiment level meas lo raises
        missing_meas_lo_config_dict = copy.deepcopy(full_lo_config_dict)
        missing_meas_lo_config_dict.pop(pulse.MeasureChannel(0))
        with self.assertRaises(QiskitError):
            qobj = assemble(self.circ, self.backend, schedule_los=missing_meas_lo_config_dict)

        # verify lo ranges are checked at experiment level
        lo_values = list(full_lo_config_dict.values())
        qubit_lo_range = [[freq - 5e6, freq + 5e6] for freq in lo_values[:5]]
        meas_lo_range = [[freq - 5e6, freq + 5e6] for freq in lo_values[5:]]

        # out of range drive lo
        full_lo_config_dict[pulse.DriveChannel(0)] -= 5.5e6
        with self.assertRaises(QiskitError):
            qobj = assemble(
                self.circ,
                self.backend,
                qubit_lo_range=qubit_lo_range,
                schedule_los=full_lo_config_dict,
            )
        full_lo_config_dict[pulse.DriveChannel(0)] += 5.5e6  # reset drive value

        # out of range meas lo
        full_lo_config_dict[pulse.MeasureChannel(0)] += 5.5e6
        with self.assertRaises(QiskitError):
            qobj = assemble(
                self.circ,
                self.backend,
                meas_lo_range=meas_lo_range,
                schedule_los=full_lo_config_dict,
            )


class TestPulseAssembler(QiskitTestCase):
    """Tests for assembling schedules to qobj."""

    def setUp(self):
        super().setUp()
        self.backend = FakeOpenPulse2Q()
        self.backend_config = self.backend.configuration()

        test_pulse = pulse.Waveform(
            samples=np.array([0.02739068, 0.05, 0.05, 0.05, 0.02739068], dtype=np.complex128),
            name="pulse0",
        )

        self.schedule = pulse.Schedule(name="fake_experiment")
        self.schedule = self.schedule.insert(0, Play(test_pulse, self.backend_config.drive(0)))
        for i in range(self.backend_config.n_qubits):
            self.schedule = self.schedule.insert(
                5, Acquire(5, self.backend_config.acquire(i), MemorySlot(i))
            )

        self.user_lo_config_dict = {self.backend_config.drive(0): 4.91e9}
        self.user_lo_config = pulse.LoConfig(self.user_lo_config_dict)

        self.default_qubit_lo_freq = [4.9e9, 5.0e9]
        self.default_meas_lo_freq = [6.5e9, 6.6e9]

        self.config = {"meas_level": 1, "memory_slot_size": 100, "meas_return": "avg"}

        self.header = {"backend_name": "FakeOpenPulse2Q", "backend_version": "0.0.0"}

    def test_assemble_adds_schedule_metadata_to_experiment_header(self):
        """Verify that any circuit metadata is added to the exeriment header."""
        self.schedule.metadata = {"experiment_type": "gst", "execution_number": "1234"}
        qobj = assemble(
            self.schedule,
            shots=100,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=[],
        )
        self.assertEqual(
            qobj.experiments[0].header.metadata,
            {"experiment_type": "gst", "execution_number": "1234"},
        )

    def test_assemble_sample_pulse(self):
        """Test that the pulse lib and qobj instruction can be paired up."""
        schedule = pulse.Schedule()
        schedule += pulse.Play(
            pulse.Waveform([0.1] * 16, name="test0"), pulse.DriveChannel(0), name="test1"
        )
        schedule += pulse.Play(
            pulse.Waveform([0.1] * 16, name="test1"), pulse.DriveChannel(0), name="test2"
        )
        schedule += pulse.Play(
            pulse.Waveform([0.5] * 16, name="test0"), pulse.DriveChannel(0), name="test1"
        )
        qobj = assemble(
            schedule,
            qobj_header=self.header,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=[],
            **self.config,
        )

        test_dict = qobj.to_dict()
        experiment = test_dict["experiments"][0]
        inst0_name = experiment["instructions"][0]["name"]
        inst1_name = experiment["instructions"][1]["name"]
        inst2_name = experiment["instructions"][2]["name"]
        pulses = {}
        for item in test_dict["config"]["pulse_library"]:
            pulses[item["name"]] = item["samples"]
        self.assertTrue(all(name in pulses for name in [inst0_name, inst1_name, inst2_name]))
        # Their pulses are the same
        self.assertEqual(inst0_name, inst1_name)
        self.assertTrue(np.allclose(pulses[inst0_name], [0.1] * 16))
        self.assertTrue(np.allclose(pulses[inst2_name], [0.5] * 16))

    def test_assemble_single_schedule_without_lo_config(self):
        """Test assembling a single schedule, no lo config."""
        qobj = assemble(
            self.schedule,
            qobj_header=self.header,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=[],
            **self.config,
        )

        test_dict = qobj.to_dict()
        self.assertListEqual(test_dict["config"]["qubit_lo_freq"], [4.9, 5.0])
        self.assertEqual(len(test_dict["experiments"]), 1)
        self.assertEqual(len(test_dict["experiments"][0]["instructions"]), 2)

    def test_assemble_multi_schedules_without_lo_config(self):
        """Test assembling schedules, no lo config."""
        qobj = assemble(
            [self.schedule, self.schedule],
            qobj_header=self.header,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            **self.config,
        )

        test_dict = qobj.to_dict()
        self.assertListEqual(test_dict["config"]["qubit_lo_freq"], [4.9, 5.0])
        self.assertEqual(len(test_dict["experiments"]), 2)
        self.assertEqual(len(test_dict["experiments"][0]["instructions"]), 2)

    def test_assemble_single_schedule_with_lo_config(self):
        """Test assembling a single schedule, with a single lo config."""
        qobj = assemble(
            self.schedule,
            qobj_header=self.header,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=self.user_lo_config,
            **self.config,
        )

        test_dict = qobj.to_dict()
        self.assertListEqual(test_dict["config"]["qubit_lo_freq"], [4.91, 5.0])
        self.assertEqual(len(test_dict["experiments"]), 1)
        self.assertEqual(len(test_dict["experiments"][0]["instructions"]), 2)

    def test_assemble_single_schedule_with_lo_config_dict(self):
        """Test assembling a single schedule, with a single lo config supplied as dictionary."""
        qobj = assemble(
            self.schedule,
            qobj_header=self.header,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=self.user_lo_config_dict,
            **self.config,
        )

        test_dict = qobj.to_dict()
        self.assertListEqual(test_dict["config"]["qubit_lo_freq"], [4.91, 5.0])
        self.assertEqual(len(test_dict["experiments"]), 1)
        self.assertEqual(len(test_dict["experiments"][0]["instructions"]), 2)

    def test_assemble_single_schedule_with_multi_lo_configs(self):
        """Test assembling a single schedule, with multiple lo configs (frequency sweep)."""
        qobj = assemble(
            self.schedule,
            qobj_header=self.header,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=[self.user_lo_config, self.user_lo_config],
            **self.config,
        )
        test_dict = qobj.to_dict()

        self.assertListEqual(test_dict["config"]["qubit_lo_freq"], [4.9, 5.0])
        self.assertEqual(len(test_dict["experiments"]), 2)
        self.assertEqual(len(test_dict["experiments"][0]["instructions"]), 2)
        self.assertDictEqual(test_dict["experiments"][0]["config"], {"qubit_lo_freq": [4.91, 5.0]})

    def test_assemble_multi_schedules_with_multi_lo_configs(self):
        """Test assembling schedules, with the same number of lo configs (n:n setup)."""
        qobj = assemble(
            [self.schedule, self.schedule],
            qobj_header=self.header,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=[self.user_lo_config, self.user_lo_config],
            **self.config,
        )

        test_dict = qobj.to_dict()
        self.assertListEqual(test_dict["config"]["qubit_lo_freq"], [4.9, 5.0])
        self.assertEqual(len(test_dict["experiments"]), 2)
        self.assertEqual(len(test_dict["experiments"][0]["instructions"]), 2)
        self.assertDictEqual(test_dict["experiments"][0]["config"], {"qubit_lo_freq": [4.91, 5.0]})

    def test_assemble_multi_schedules_with_wrong_number_of_multi_lo_configs(self):
        """Test assembling schedules, with a different number of lo configs (n:m setup)."""
        with self.assertRaises(QiskitError):
            assemble(
                [self.schedule, self.schedule, self.schedule],
                qobj_header=self.header,
                qubit_lo_freq=self.default_qubit_lo_freq,
                meas_lo_freq=self.default_meas_lo_freq,
                schedule_los=[self.user_lo_config, self.user_lo_config],
                **self.config,
            )

    def test_assemble_meas_map(self):
        """Test assembling a single schedule, no lo config."""
        schedule = Schedule(name="fake_experiment")
        schedule = schedule.insert(5, Acquire(5, AcquireChannel(0), MemorySlot(0)))
        schedule = schedule.insert(5, Acquire(5, AcquireChannel(1), MemorySlot(1)))

        qobj = assemble(
            schedule,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            meas_map=[[0], [1]],
        )
        self.assertIsInstance(qobj, PulseQobj)

        qobj = assemble(
            schedule,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            meas_map=[[0, 1, 2]],
        )
        self.assertIsInstance(qobj, PulseQobj)

    def test_assemble_memory_slots(self):
        """Test assembling a schedule and inferring number of memoryslots."""
        n_memoryslots = 10

        # single acquisition
        schedule = Acquire(
            5, self.backend_config.acquire(0), mem_slot=pulse.MemorySlot(n_memoryslots - 1)
        )

        qobj = assemble(
            schedule,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            meas_map=[[0], [1]],
        )

        self.assertEqual(qobj.config.memory_slots, n_memoryslots)
        # this should be in experimental header as well
        self.assertEqual(qobj.experiments[0].header.memory_slots, n_memoryslots)

        # multiple acquisition
        schedule = Acquire(
            5, self.backend_config.acquire(0), mem_slot=pulse.MemorySlot(n_memoryslots - 1)
        )
        schedule = schedule.insert(
            10,
            Acquire(
                5, self.backend_config.acquire(0), mem_slot=pulse.MemorySlot(n_memoryslots - 1)
            ),
        )

        qobj = assemble(
            schedule,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            meas_map=[[0], [1]],
        )

        self.assertEqual(qobj.config.memory_slots, n_memoryslots)
        # this should be in experimental header as well
        self.assertEqual(qobj.experiments[0].header.memory_slots, n_memoryslots)

    def test_assemble_memory_slots_for_schedules(self):
        """Test assembling schedules with different memory slots."""
        n_memoryslots = [10, 5, 7]

        schedules = []
        for n_memoryslot in n_memoryslots:
            schedule = Acquire(
                5, self.backend_config.acquire(0), mem_slot=pulse.MemorySlot(n_memoryslot - 1)
            )
            schedules.append(schedule)

        qobj = assemble(
            schedules,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            meas_map=[[0], [1]],
        )

        self.assertEqual(qobj.config.memory_slots, max(n_memoryslots))
        self.assertEqual(qobj.experiments[0].header.memory_slots, n_memoryslots[0])
        self.assertEqual(qobj.experiments[1].header.memory_slots, n_memoryslots[1])
        self.assertEqual(qobj.experiments[2].header.memory_slots, n_memoryslots[2])

    def test_pulse_name_conflicts(self):
        """Test that pulse name conflicts can be resolved."""
        name_conflict_pulse = pulse.Waveform(
            samples=np.array([0.02, 0.05, 0.05, 0.05, 0.02], dtype=np.complex128), name="pulse0"
        )

        self.schedule = self.schedule.insert(
            1, Play(name_conflict_pulse, self.backend_config.drive(1))
        )

        qobj = assemble(
            self.schedule,
            qobj_header=self.header,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=[],
            **self.config,
        )

        self.assertNotEqual(qobj.config.pulse_library[0].name, qobj.config.pulse_library[1].name)

    def test_pulse_name_conflicts_in_other_schedule(self):
        """Test two pulses with the same name in different schedule can be resolved."""
        backend = FakeHanoi()
        defaults = backend.defaults()

        schedules = []
        ch_d0 = pulse.DriveChannel(0)
        for amp in (0.1, 0.2):
            sched = Schedule()
            sched += Play(gaussian(duration=100, amp=amp, sigma=30, name="my_pulse"), ch_d0)
            sched += measure(qubits=[0], backend=backend) << 100
            schedules.append(sched)

        qobj = assemble(
            schedules, qubit_lo_freq=defaults.qubit_freq_est, meas_lo_freq=defaults.meas_freq_est
        )

        # two user pulses and one measurement pulse should be contained
        self.assertEqual(len(qobj.config.pulse_library), 3)

    def test_assemble_with_delay(self):
        """Test that delay instruction is not ignored in assembly."""
        delay_schedule = pulse.Delay(10, self.backend_config.drive(0))
        delay_schedule += self.schedule
        delay_qobj = assemble(delay_schedule, self.backend)

        self.assertEqual(delay_qobj.experiments[0].instructions[0].name, "delay")
        self.assertEqual(delay_qobj.experiments[0].instructions[0].duration, 10)
        self.assertEqual(delay_qobj.experiments[0].instructions[0].t0, 0)

    def test_delay_removed_on_acq_ch(self):
        """Test that delay instructions on acquire channels are skipped on assembly with times
        shifted properly.
        """
        delay0 = pulse.Delay(5, self.backend_config.acquire(0))
        delay1 = pulse.Delay(7, self.backend_config.acquire(1))

        sched0 = delay0
        sched0 += self.schedule  # includes ``Acquire`` instr
        sched0 += delay1

        sched1 = self.schedule  # includes ``Acquire`` instr
        sched1 += delay0
        sched1 += delay1

        sched2 = delay0
        sched2 += delay1
        sched2 += self.schedule  # includes ``Acquire`` instr

        delay_qobj = assemble([sched0, sched1, sched2], self.backend)

        # check that no delay instrs occur on acquire channels
        is_acq_delay = False
        for exp in delay_qobj.experiments:
            for instr in exp.instructions:
                if instr.name == "delay" and "a" in instr.ch:
                    is_acq_delay = True

        self.assertFalse(is_acq_delay)

        # check that acquire instr are shifted from ``t0=5`` as needed
        self.assertEqual(delay_qobj.experiments[0].instructions[1].t0, 10)
        self.assertEqual(delay_qobj.experiments[0].instructions[1].name, "acquire")
        self.assertEqual(delay_qobj.experiments[1].instructions[1].t0, 5)
        self.assertEqual(delay_qobj.experiments[1].instructions[1].name, "acquire")
        self.assertEqual(delay_qobj.experiments[2].instructions[1].t0, 12)
        self.assertEqual(delay_qobj.experiments[2].instructions[1].name, "acquire")

    def test_assemble_schedule_enum(self):
        """Test assembling a schedule with enum input values to assemble."""
        qobj = assemble(
            self.schedule,
            qobj_header=self.header,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            schedule_los=[],
            meas_level=MeasLevel.CLASSIFIED,
            meas_return=MeasReturnType.AVERAGE,
        )

        test_dict = qobj.to_dict()
        self.assertEqual(test_dict["config"]["meas_return"], "avg")
        self.assertEqual(test_dict["config"]["meas_level"], 2)

    def test_assemble_parametric(self):
        """Test that parametric pulses can be assembled properly into a PulseQobj."""
        sched = pulse.Schedule(name="test_parametric")
        sched += Play(pulse.Gaussian(duration=25, sigma=4, amp=0.5j), DriveChannel(0))
        sched += Play(pulse.Drag(duration=25, amp=0.2 + 0.3j, sigma=7.8, beta=4), DriveChannel(1))
        sched += Play(pulse.Constant(duration=25, amp=1), DriveChannel(2))
        sched += (
            Play(pulse.GaussianSquare(duration=150, amp=0.2, sigma=8, width=140), MeasureChannel(0))
            << sched.duration
        )
        backend = FakeOpenPulse3Q()
        backend.configuration().parametric_pulses = [
            "gaussian",
            "drag",
            "gaussian_square",
            "constant",
        ]
        qobj = assemble(sched, backend)

        self.assertEqual(qobj.config.pulse_library, [])
        qobj_insts = qobj.experiments[0].instructions
        self.assertTrue(all(inst.name == "parametric_pulse" for inst in qobj_insts))
        self.assertEqual(qobj_insts[0].pulse_shape, "gaussian")
        self.assertEqual(qobj_insts[1].pulse_shape, "drag")
        self.assertEqual(qobj_insts[2].pulse_shape, "constant")
        self.assertEqual(qobj_insts[3].pulse_shape, "gaussian_square")
        self.assertDictEqual(qobj_insts[0].parameters, {"duration": 25, "sigma": 4, "amp": 0.5j})
        self.assertDictEqual(
            qobj_insts[1].parameters, {"duration": 25, "sigma": 7.8, "amp": 0.2 + 0.3j, "beta": 4}
        )
        self.assertDictEqual(qobj_insts[2].parameters, {"duration": 25, "amp": 1})
        self.assertDictEqual(
            qobj_insts[3].parameters, {"duration": 150, "sigma": 8, "amp": 0.2, "width": 140}
        )
        self.assertEqual(
            qobj.to_dict()["experiments"][0]["instructions"][0]["parameters"]["amp"], 0.5j
        )

    def test_assemble_parametric_unsupported(self):
        """Test that parametric pulses are translated to Waveform if they're not supported
        by the backend during assemble time.
        """
        sched = pulse.Schedule(name="test_parametric_to_sample_pulse")
        sched += Play(pulse.Drag(duration=25, amp=0.2 + 0.3j, sigma=7.8, beta=4), DriveChannel(1))
        sched += Play(pulse.Constant(duration=25, amp=1), DriveChannel(2))

        backend = FakeOpenPulse3Q()
        backend.configuration().parametric_pulses = ["something_extra"]

        qobj = assemble(sched, backend)

        self.assertNotEqual(qobj.config.pulse_library, [])
        qobj_insts = qobj.experiments[0].instructions
        self.assertFalse(hasattr(qobj_insts[0], "pulse_shape"))

    def test_assemble_parametric_pulse_kwarg_with_backend_setting(self):
        """Test that parametric pulses respect the kwarg over backend"""
        backend = FakeHanoi()

        qc = QuantumCircuit(1, 1)
        qc.x(0)
        qc.measure(0, 0)
        with pulse.build(backend, name="x") as x_q0:
            pulse.play(pulse.Gaussian(duration=128, amp=0.1, sigma=16), pulse.drive_channel(0))

        qc.add_calibration("x", (0,), x_q0)

        qobj = assemble(qc, backend, parametric_pulses=["gaussian"])
        self.assertEqual(qobj.config.parametric_pulses, ["gaussian"])

    def test_assemble_parametric_pulse_kwarg_empty_list_with_backend_setting(self):
        """Test that parametric pulses respect the kwarg as empty list over backend"""
        backend = FakeHanoi()

        qc = QuantumCircuit(1, 1)
        qc.x(0)
        qc.measure(0, 0)
        with pulse.build(backend, name="x") as x_q0:
            pulse.play(pulse.Gaussian(duration=128, amp=0.1, sigma=16), pulse.drive_channel(0))

        qc.add_calibration("x", (0,), x_q0)

        qobj = assemble(qc, backend, parametric_pulses=[])
        self.assertEqual(qobj.config.parametric_pulses, [])

    def test_init_qubits_default(self):
        """Check that the init_qubits=None assemble option is passed on to the qobj."""
        qobj = assemble(self.schedule, self.backend)
        self.assertEqual(qobj.config.init_qubits, True)

    def test_init_qubits_true(self):
        """Check that the init_qubits=True assemble option is passed on to the qobj."""
        qobj = assemble(self.schedule, self.backend, init_qubits=True)
        self.assertEqual(qobj.config.init_qubits, True)

    def test_init_qubits_false(self):
        """Check that the init_qubits=False assemble option is passed on to the qobj."""
        qobj = assemble(self.schedule, self.backend, init_qubits=False)
        self.assertEqual(qobj.config.init_qubits, False)

    def test_assemble_backend_rep_times_delays(self):
        """Check that rep_time and rep_delay are properly set from backend values."""
        # use first entry from allowed backend values
        rep_times = [2.0, 3.0, 4.0]  # sec
        rep_delay_range = [2.5e-3, 4.5e-3]
        default_rep_delay = 3.0e-3
        self.backend_config.rep_times = rep_times
        setattr(self.backend_config, "rep_delay_range", rep_delay_range)
        setattr(self.backend_config, "default_rep_delay", default_rep_delay)

        # dynamic rep rates off
        qobj = assemble(self.schedule, self.backend)
        self.assertEqual(qobj.config.rep_time, int(rep_times[0] * 1e6))
        self.assertEqual(hasattr(qobj.config, "rep_delay"), False)

        # dynamic rep rates on
        setattr(self.backend_config, "dynamic_reprate_enabled", True)
        # RuntimeWarning bc ``rep_time`` is specified`` when dynamic rep rates not enabled
        with self.assertWarns(RuntimeWarning):
            qobj = assemble(self.schedule, self.backend)
        self.assertEqual(qobj.config.rep_time, int(rep_times[0] * 1e6))
        self.assertEqual(qobj.config.rep_delay, default_rep_delay * 1e6)

    def test_assemble_user_rep_time_delay(self):
        """Check that user runtime config rep_time and rep_delay work."""
        # set custom rep_time and rep_delay in runtime config
        rep_time = 200.0e-6
        rep_delay = 2.5e-6
        self.config["rep_time"] = rep_time
        self.config["rep_delay"] = rep_delay

        # dynamic rep rates off
        # RuntimeWarning bc using ``rep_delay`` when dynamic rep rates off
        with self.assertWarns(RuntimeWarning):
            qobj = assemble(self.schedule, self.backend, **self.config)
        self.assertEqual(qobj.config.rep_time, int(rep_time * 1e6))
        self.assertEqual(hasattr(qobj.config, "rep_delay"), False)

        # now remove rep_delay and enable dynamic rep rates
        # RuntimeWarning bc using ``rep_time`` when dynamic rep rates are enabled
        del self.config["rep_delay"]
        setattr(self.backend_config, "dynamic_reprate_enabled", True)
        with self.assertWarns(RuntimeWarning):
            qobj = assemble(self.schedule, self.backend, **self.config)
        self.assertEqual(qobj.config.rep_time, int(rep_time * 1e6))
        self.assertEqual(hasattr(qobj.config, "rep_delay"), False)

        # use ``default_rep_delay``
        # ``rep_time`` comes from allowed backend rep_times
        rep_times = [0.5, 1.0, 1.5]  # sec
        self.backend_config.rep_times = rep_times
        setattr(self.backend_config, "rep_delay_range", [0, 3.0e-6])
        setattr(self.backend_config, "default_rep_delay", 2.2e-6)
        del self.config["rep_time"]
        qobj = assemble(self.schedule, self.backend, **self.config)
        self.assertEqual(qobj.config.rep_time, int(rep_times[0] * 1e6))
        self.assertEqual(qobj.config.rep_delay, 2.2)

        # use qobj ``default_rep_delay``
        self.config["rep_delay"] = 1.5e-6
        qobj = assemble(self.schedule, self.backend, **self.config)
        self.assertEqual(qobj.config.rep_time, int(rep_times[0] * 1e6))
        self.assertEqual(qobj.config.rep_delay, 1.5)

        # use ``rep_delay`` outside of ``rep_delay_range
        self.config["rep_delay"] = 5.0e-6
        with self.assertRaises(QiskitError):
            assemble(self.schedule, self.backend, **self.config)

    def test_assemble_with_individual_discriminators(self):
        """Test that assembly works with individual discriminators."""
        disc_one = Discriminator("disc_one", test_params=True)
        disc_two = Discriminator("disc_two", test_params=False)

        schedule = Schedule()
        schedule = schedule.append(
            Acquire(5, AcquireChannel(0), MemorySlot(0), discriminator=disc_one),
        )
        schedule = schedule.append(
            Acquire(5, AcquireChannel(1), MemorySlot(1), discriminator=disc_two),
        )

        qobj = assemble(
            schedule,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            meas_map=[[0, 1]],
        )

        qobj_discriminators = qobj.experiments[0].instructions[0].discriminators
        self.assertEqual(len(qobj_discriminators), 2)
        self.assertEqual(qobj_discriminators[0].name, "disc_one")
        self.assertEqual(qobj_discriminators[0].params["test_params"], True)
        self.assertEqual(qobj_discriminators[1].name, "disc_two")
        self.assertEqual(qobj_discriminators[1].params["test_params"], False)

    def test_assemble_with_single_discriminators(self):
        """Test that assembly works with both a single discriminator."""
        disc_one = Discriminator("disc_one", test_params=True)

        schedule = Schedule()
        schedule = schedule.append(
            Acquire(5, AcquireChannel(0), MemorySlot(0), discriminator=disc_one),
        )
        schedule = schedule.append(
            Acquire(5, AcquireChannel(1), MemorySlot(1)),
        )

        qobj = assemble(
            schedule,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            meas_map=[[0, 1]],
        )

        qobj_discriminators = qobj.experiments[0].instructions[0].discriminators
        self.assertEqual(len(qobj_discriminators), 1)
        self.assertEqual(qobj_discriminators[0].name, "disc_one")
        self.assertEqual(qobj_discriminators[0].params["test_params"], True)

    def test_assemble_with_unequal_discriminators(self):
        """Test that assembly works with incorrect number of discriminators for
        number of qubits."""
        disc_one = Discriminator("disc_one", test_params=True)
        disc_two = Discriminator("disc_two", test_params=False)

        schedule = Schedule()
        schedule += Acquire(5, AcquireChannel(0), MemorySlot(0), discriminator=disc_one)
        schedule += Acquire(5, AcquireChannel(1), MemorySlot(1), discriminator=disc_two)
        schedule += Acquire(5, AcquireChannel(2), MemorySlot(2))

        with self.assertRaises(QiskitError):
            assemble(
                schedule,
                qubit_lo_freq=self.default_qubit_lo_freq,
                meas_lo_freq=self.default_meas_lo_freq,
                meas_map=[[0, 1, 2]],
            )

    def test_assemble_with_individual_kernels(self):
        """Test that assembly works with individual kernels."""
        disc_one = Kernel("disc_one", test_params=True)
        disc_two = Kernel("disc_two", test_params=False)

        schedule = Schedule()
        schedule = schedule.append(
            Acquire(5, AcquireChannel(0), MemorySlot(0), kernel=disc_one),
        )
        schedule = schedule.append(
            Acquire(5, AcquireChannel(1), MemorySlot(1), kernel=disc_two),
        )

        qobj = assemble(
            schedule,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            meas_map=[[0, 1]],
        )

        qobj_kernels = qobj.experiments[0].instructions[0].kernels
        self.assertEqual(len(qobj_kernels), 2)
        self.assertEqual(qobj_kernels[0].name, "disc_one")
        self.assertEqual(qobj_kernels[0].params["test_params"], True)
        self.assertEqual(qobj_kernels[1].name, "disc_two")
        self.assertEqual(qobj_kernels[1].params["test_params"], False)

    def test_assemble_with_single_kernels(self):
        """Test that assembly works with both a single kernel."""
        disc_one = Kernel("disc_one", test_params=True)

        schedule = Schedule()
        schedule = schedule.append(
            Acquire(5, AcquireChannel(0), MemorySlot(0), kernel=disc_one),
        )
        schedule = schedule.append(
            Acquire(5, AcquireChannel(1), MemorySlot(1)),
        )

        qobj = assemble(
            schedule,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            meas_map=[[0, 1]],
        )

        qobj_kernels = qobj.experiments[0].instructions[0].kernels
        self.assertEqual(len(qobj_kernels), 1)
        self.assertEqual(qobj_kernels[0].name, "disc_one")
        self.assertEqual(qobj_kernels[0].params["test_params"], True)

    def test_assemble_with_unequal_kernels(self):
        """Test that assembly works with incorrect number of discriminators for
        number of qubits."""
        disc_one = Kernel("disc_one", test_params=True)
        disc_two = Kernel("disc_two", test_params=False)

        schedule = Schedule()
        schedule += Acquire(5, AcquireChannel(0), MemorySlot(0), kernel=disc_one)
        schedule += Acquire(5, AcquireChannel(1), MemorySlot(1), kernel=disc_two)
        schedule += Acquire(5, AcquireChannel(2), MemorySlot(2))

        with self.assertRaises(QiskitError):
            assemble(
                schedule,
                qubit_lo_freq=self.default_qubit_lo_freq,
                meas_lo_freq=self.default_meas_lo_freq,
                meas_map=[[0, 1, 2]],
            )

    def test_assemble_single_instruction(self):
        """Test assembling schedules, no lo config."""
        inst = pulse.Play(pulse.Constant(100, 1.0), pulse.DriveChannel(0))
        self.assertIsInstance(assemble(inst, self.backend), PulseQobj)

    def test_assemble_overlapping_time(self):
        """Test that assembly errors when qubits are measured in overlapping time."""
        schedule = Schedule()
        schedule = schedule.append(
            Acquire(5, AcquireChannel(0), MemorySlot(0)),
        )
        schedule = schedule.append(
            Acquire(5, AcquireChannel(1), MemorySlot(1)) << 1,
        )
        with self.assertRaises(QiskitError):
            assemble(
                schedule,
                qubit_lo_freq=self.default_qubit_lo_freq,
                meas_lo_freq=self.default_meas_lo_freq,
                meas_map=[[0, 1]],
            )

    def test_assemble_meas_map_vs_insts(self):
        """Test that assembly errors when the qubits are measured in overlapping time
        and qubits are not in the first meas_map list."""
        schedule = Schedule()
        schedule += Acquire(5, AcquireChannel(0), MemorySlot(0))
        schedule += Acquire(5, AcquireChannel(1), MemorySlot(1))
        schedule += Acquire(5, AcquireChannel(2), MemorySlot(2)) << 2
        schedule += Acquire(5, AcquireChannel(3), MemorySlot(3)) << 2

        with self.assertRaises(QiskitError):
            assemble(
                schedule,
                qubit_lo_freq=self.default_qubit_lo_freq,
                meas_lo_freq=self.default_meas_lo_freq,
                meas_map=[[0], [1, 2], [3]],
            )

    def test_assemble_non_overlapping_time_single_meas_map(self):
        """Test that assembly works when qubits are measured in non-overlapping
        time within the same measurement map list."""
        schedule = Schedule()
        schedule = schedule.append(
            Acquire(5, AcquireChannel(0), MemorySlot(0)),
        )
        schedule = schedule.append(
            Acquire(5, AcquireChannel(1), MemorySlot(1)) << 5,
        )
        qobj = assemble(
            schedule,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            meas_map=[[0, 1]],
        )
        self.assertIsInstance(qobj, PulseQobj)

    def test_assemble_disjoint_time(self):
        """Test that assembly works when qubits are in disjoint meas map sets."""
        schedule = Schedule()
        schedule = schedule.append(
            Acquire(5, AcquireChannel(0), MemorySlot(0)),
        )
        schedule = schedule.append(
            Acquire(5, AcquireChannel(1), MemorySlot(1)) << 1,
        )
        qobj = assemble(
            schedule,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            meas_map=[[0, 2], [1, 3]],
        )
        self.assertIsInstance(qobj, PulseQobj)

    def test_assemble_valid_qubits(self):
        """Test that assembly works when qubits that are in the measurement map
        is measured."""
        schedule = Schedule()
        schedule = schedule.append(
            Acquire(5, AcquireChannel(1), MemorySlot(1)),
        )
        schedule = schedule.append(
            Acquire(5, AcquireChannel(2), MemorySlot(2)),
        )
        schedule = schedule.append(
            Acquire(5, AcquireChannel(3), MemorySlot(3)),
        )
        qobj = assemble(
            schedule,
            qubit_lo_freq=self.default_qubit_lo_freq,
            meas_lo_freq=self.default_meas_lo_freq,
            meas_map=[[0, 1, 2], [3]],
        )
        self.assertIsInstance(qobj, PulseQobj)


class TestPulseAssemblerMissingKwargs(QiskitTestCase):
    """Verify that errors are raised in case backend is not provided and kwargs are missing."""

    def setUp(self):
        super().setUp()
        self.schedule = pulse.Schedule(name="fake_experiment")

        self.backend = FakeOpenPulse2Q()
        self.config = self.backend.configuration()
        self.defaults = self.backend.defaults()
        self.qubit_lo_freq = list(self.defaults.qubit_freq_est)
        self.meas_lo_freq = list(self.defaults.meas_freq_est)
        self.qubit_lo_range = self.config.qubit_lo_range
        self.meas_lo_range = self.config.meas_lo_range
        self.schedule_los = {
            pulse.DriveChannel(0): self.qubit_lo_freq[0],
            pulse.DriveChannel(1): self.qubit_lo_freq[1],
            pulse.MeasureChannel(0): self.meas_lo_freq[0],
            pulse.MeasureChannel(1): self.meas_lo_freq[1],
        }
        self.meas_map = self.config.meas_map
        self.memory_slots = self.config.n_qubits

        # default rep_time and rep_delay
        self.rep_time = self.config.rep_times[0]
        self.rep_delay = None

    def test_defaults(self):
        """Test defaults work."""
        qobj = assemble(
            self.schedule,
            qubit_lo_freq=self.qubit_lo_freq,
            meas_lo_freq=self.meas_lo_freq,
            qubit_lo_range=self.qubit_lo_range,
            meas_lo_range=self.meas_lo_range,
            schedule_los=self.schedule_los,
            meas_map=self.meas_map,
            memory_slots=self.memory_slots,
            rep_time=self.rep_time,
            rep_delay=self.rep_delay,
        )
        self.assertIsInstance(qobj, PulseQobj)

    def test_missing_qubit_lo_freq(self):
        """Test error raised if qubit_lo_freq missing."""

        with self.assertRaises(QiskitError):
            assemble(
                self.schedule,
                qubit_lo_freq=None,
                meas_lo_freq=self.meas_lo_freq,
                qubit_lo_range=self.qubit_lo_range,
                meas_lo_range=self.meas_lo_range,
                meas_map=self.meas_map,
                memory_slots=self.memory_slots,
                rep_time=self.rep_time,
                rep_delay=self.rep_delay,
            )

    def test_missing_meas_lo_freq(self):
        """Test error raised if meas_lo_freq missing."""

        with self.assertRaises(QiskitError):
            assemble(
                self.schedule,
                qubit_lo_freq=self.qubit_lo_freq,
                meas_lo_freq=None,
                qubit_lo_range=self.qubit_lo_range,
                meas_lo_range=self.meas_lo_range,
                meas_map=self.meas_map,
                memory_slots=self.memory_slots,
                rep_time=self.rep_time,
                rep_delay=self.rep_delay,
            )

    def test_missing_memory_slots(self):
        """Test error is not raised if memory_slots are missing."""
        qobj = assemble(
            self.schedule,
            qubit_lo_freq=self.qubit_lo_freq,
            meas_lo_freq=self.meas_lo_freq,
            qubit_lo_range=self.qubit_lo_range,
            meas_lo_range=self.meas_lo_range,
            schedule_los=self.schedule_los,
            meas_map=self.meas_map,
            memory_slots=None,
            rep_time=self.rep_time,
            rep_delay=self.rep_delay,
        )
        self.assertIsInstance(qobj, PulseQobj)

    def test_missing_rep_time_and_delay(self):
        """Test qobj is valid if rep_time and rep_delay are missing."""
        qobj = assemble(
            self.schedule,
            qubit_lo_freq=self.qubit_lo_freq,
            meas_lo_freq=self.meas_lo_freq,
            qubit_lo_range=self.qubit_lo_range,
            meas_lo_range=self.meas_lo_range,
            schedule_los=self.schedule_los,
            meas_map=self.meas_map,
            memory_slots=None,
            rep_time=None,
            rep_delay=None,
        )
        self.assertEqual(hasattr(qobj, "rep_time"), False)
        self.assertEqual(hasattr(qobj, "rep_delay"), False)

    def test_missing_meas_map(self):
        """Test that assembly still works if meas_map is missing."""
        qobj = assemble(
            self.schedule,
            qubit_lo_freq=self.qubit_lo_freq,
            meas_lo_freq=self.meas_lo_freq,
            qubit_lo_range=self.qubit_lo_range,
            meas_lo_range=self.meas_lo_range,
            schedule_los=self.schedule_los,
            meas_map=None,
            memory_slots=self.memory_slots,
            rep_time=self.rep_time,
            rep_delay=self.rep_delay,
        )
        self.assertIsInstance(qobj, PulseQobj)

    def test_missing_lo_ranges(self):
        """Test that assembly still works if lo_ranges are missing."""
        qobj = assemble(
            self.schedule,
            qubit_lo_freq=self.qubit_lo_freq,
            meas_lo_freq=self.meas_lo_freq,
            qubit_lo_range=None,
            meas_lo_range=None,
            schedule_los=self.schedule_los,
            meas_map=self.meas_map,
            memory_slots=self.memory_slots,
            rep_time=self.rep_time,
            rep_delay=self.rep_delay,
        )
        self.assertIsInstance(qobj, PulseQobj)

    def test_unsupported_meas_level(self):
        """Test that assembly raises an error if meas_level is not supported"""
        backend = FakeOpenPulse2Q()
        backend.configuration().meas_levels = [1, 2]
        with self.assertRaises(QiskitError):
            assemble(
                self.schedule,
                backend,
                qubit_lo_freq=self.qubit_lo_freq,
                meas_lo_freq=self.meas_lo_freq,
                qubit_lo_range=self.qubit_lo_range,
                meas_lo_range=self.meas_lo_range,
                schedule_los=self.schedule_los,
                meas_level=0,
                meas_map=self.meas_map,
                memory_slots=self.memory_slots,
                rep_time=self.rep_time,
                rep_delay=self.rep_delay,
            )

    def test_single_and_deprecated_acquire_styles(self):
        """Test that acquires are identically combined with Acquires that take a single channel."""
        backend = FakeOpenPulse2Q()
        new_style_schedule = Schedule()
        acq_dur = 1200
        for i in range(2):
            new_style_schedule += Acquire(acq_dur, AcquireChannel(i), MemorySlot(i))

        deprecated_style_schedule = Schedule()
        for i in range(2):
            deprecated_style_schedule += Acquire(1200, AcquireChannel(i), MemorySlot(i))

        # The Qobj IDs will be different
        n_qobj = assemble(new_style_schedule, backend)
        n_qobj.qobj_id = None
        n_qobj.experiments[0].header.name = None
        d_qobj = assemble(deprecated_style_schedule, backend)
        d_qobj.qobj_id = None
        d_qobj.experiments[0].header.name = None
        self.assertEqual(n_qobj, d_qobj)

        assembled_acquire = n_qobj.experiments[0].instructions[0]
        self.assertEqual(assembled_acquire.qubits, [0, 1])
        self.assertEqual(assembled_acquire.memory_slot, [0, 1])


class StreamHandlerRaiseException(StreamHandler):
    """Handler class that will raise an exception on formatting errors."""

    def handleError(self, record):
        raise sys.exc_info()


class TestLogAssembler(QiskitTestCase):
    """Testing the log_assembly option."""

    def setUp(self):
        super().setUp()
        logger = getLogger()
        self.addCleanup(logger.setLevel, logger.level)
        logger.setLevel("DEBUG")
        self.output = io.StringIO()
        logger.addHandler(StreamHandlerRaiseException(self.output))
        self.circuit = QuantumCircuit(QuantumRegister(1))

    def assertAssembleLog(self, log_msg):
        """Runs assemble and checks for logs containing specified message"""
        assemble(self.circuit, shots=2000, memory=True)
        self.output.seek(0)
        # Filter unrelated log lines
        output_lines = self.output.readlines()
        assembly_log_lines = [x for x in output_lines if log_msg in x]
        self.assertTrue(len(assembly_log_lines) == 1)

    def test_assembly_log_time(self):
        """Check Total Assembly Time is logged"""
        self.assertAssembleLog("Total Assembly Time")


if __name__ == "__main__":
    unittest.main(verbosity=2)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""Implementations of boolean logic quantum circuits."""

from typing import List, Optional

from qiskit.circuit import QuantumRegister, QuantumCircuit, AncillaRegister
from qiskit.circuit.library.standard_gates import MCXGate


class AND(QuantumCircuit):
    r"""A circuit implementing the logical AND operation on a number of qubits.

    For the AND operation the state :math:`|1\rangle` is interpreted as ``True``. The result
    qubit is flipped, if the state of all variable qubits is ``True``. In this format, the AND
    operation equals a multi-controlled X gate, which is controlled on all variable qubits.
    Using a list of flags however, qubits can be skipped or negated. Practically, the flags
    allow to skip controls or to apply pre- and post-X gates to the negated qubits.

    The AND gate without special flags equals the multi-controlled-X gate:

    .. jupyter-execute::
        :hide-code:

        from qiskit.circuit.library import AND
        import qiskit.tools.jupyter
        circuit = AND(5)
        %circuit_library_info circuit

    Using flags we can negate qubits or skip them. For instance, if we have 5 qubits and want to
    return ``True`` if the first qubit is ``False`` and the last two are ``True`` we use the flags
    ``[-1, 0, 0, 1, 1]``.

    .. jupyter-execute::
        :hide-code:

        from qiskit.circuit.library import AND
        import qiskit.tools.jupyter
        circuit = AND(5, flags=[-1, 0, 0, 1, 1])
        %circuit_library_info circuit

    """

    def __init__(
        self,
        num_variable_qubits: int,
        flags: Optional[List[int]] = None,
        mcx_mode: str = "noancilla",
    ) -> None:
        """Create a new logical AND circuit.

        Args:
            num_variable_qubits: The qubits of which the OR is computed. The result will be written
                into an additional result qubit.
            flags: A list of +1/0/-1 marking negations or omissions of qubits.
            mcx_mode: The mode to be used to implement the multi-controlled X gate.
        """
        self.num_variable_qubits = num_variable_qubits
        self.flags = flags

        # add registers
        qr_variable = QuantumRegister(num_variable_qubits, name="variable")
        qr_result = QuantumRegister(1, name="result")

        circuit = QuantumCircuit(qr_variable, qr_result, name="and")

        # determine the control qubits: all that have a nonzero flag
        flags = flags or [1] * num_variable_qubits
        control_qubits = [q for q, flag in zip(qr_variable, flags) if flag != 0]

        # determine the qubits that need to be flipped (if a flag is < 0)
        flip_qubits = [q for q, flag in zip(qr_variable, flags) if flag < 0]

        # determine the number of ancillas
        num_ancillas = MCXGate.get_num_ancilla_qubits(len(control_qubits), mode=mcx_mode)
        if num_ancillas > 0:
            qr_ancilla = AncillaRegister(num_ancillas, "ancilla")
            circuit.add_register(qr_ancilla)
        else:
            qr_ancilla = []

        if len(flip_qubits) > 0:
            circuit.x(flip_qubits)
        circuit.mcx(control_qubits, qr_result[:], qr_ancilla[:], mode=mcx_mode)
        if len(flip_qubits) > 0:
            circuit.x(flip_qubits)

        super().__init__(*circuit.qregs, name="and")
        self.compose(circuit.to_gate(), qubits=self.qubits, inplace=True)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
======================================================
Executing Experiments (:mod:`qiskit.execute_function`)
======================================================

.. currentmodule:: qiskit.execute_function

.. autofunction:: execute
"""
import logging
from time import time
import warnings
from qiskit.compiler import transpile, schedule
from qiskit.providers.backend import Backend
from qiskit.pulse import Schedule, ScheduleBlock
from qiskit.exceptions import QiskitError

logger = logging.getLogger(__name__)


def _log_submission_time(start_time, end_time):
    log_msg = "Total Job Submission Time - %.5f (ms)" % ((end_time - start_time) * 1000)
    logger.info(log_msg)


def execute(
    experiments,
    backend,
    basis_gates=None,
    coupling_map=None,  # circuit transpile options
    backend_properties=None,
    initial_layout=None,
    seed_transpiler=None,
    optimization_level=None,
    pass_manager=None,
    qobj_id=None,
    qobj_header=None,
    shots=None,  # common run options
    memory=None,
    max_credits=None,
    seed_simulator=None,
    default_qubit_los=None,
    default_meas_los=None,  # schedule run options
    qubit_lo_range=None,
    meas_lo_range=None,
    schedule_los=None,
    meas_level=None,
    meas_return=None,
    memory_slots=None,
    memory_slot_size=None,
    rep_time=None,
    rep_delay=None,
    parameter_binds=None,
    schedule_circuit=False,
    inst_map=None,
    meas_map=None,
    scheduling_method=None,
    init_qubits=None,
    **run_config,
):
    """Execute a list of :class:`qiskit.circuit.QuantumCircuit` or
    :class:`qiskit.pulse.Schedule` on a backend.

    The execution is asynchronous, and a handle to a job instance is returned.

    Args:
        experiments (QuantumCircuit or list[QuantumCircuit] or Schedule or list[Schedule]):
            Circuit(s) or pulse schedule(s) to execute

        backend (Backend):
            Backend to execute circuits on.
            Transpiler options are automatically grabbed from
            backend.configuration() and backend.properties().
            If any other option is explicitly set (e.g. coupling_map), it
            will override the backend's.

        basis_gates (list[str]):
            List of basis gate names to unroll to.
            e.g: ``['u1', 'u2', 'u3', 'cx']``
            If ``None``, do not unroll.

        coupling_map (CouplingMap or list): Coupling map (perhaps custom) to
            target in mapping. Multiple formats are supported:

            #. CouplingMap instance
            #. list
               Must be given as an adjacency matrix, where each entry
               specifies all two-qubit interactions supported by backend
               e.g:
               ``[[0, 1], [0, 3], [1, 2], [1, 5], [2, 5], [4, 1], [5, 3]]``

        backend_properties (BackendProperties):
            Properties returned by a backend, including information on gate
            errors, readout errors, qubit coherence times, etc. Find a backend
            that provides this information with:
            ``backend.properties()``

        initial_layout (Layout or dict or list):
            Initial position of virtual qubits on physical qubits.
            If this layout makes the circuit compatible with the coupling_map
            constraints, it will be used.
            The final layout is not guaranteed to be the same, as the transpiler
            may permute qubits through swaps or other means.

            Multiple formats are supported:

            #. :class:`qiskit.transpiler.Layout` instance
            #. ``dict``:
               virtual to physical::

                    {qr[0]: 0,
                     qr[1]: 3,
                     qr[2]: 5}

               physical to virtual::
                    {0: qr[0],
                     3: qr[1],
                     5: qr[2]}

            #. ``list``
               virtual to physical::

                    [0, 3, 5]  # virtual qubits are ordered (in addition to named)

               physical to virtual::

                    [qr[0], None, None, qr[1], None, qr[2]]

        seed_transpiler (int): Sets random seed for the stochastic parts of the transpiler

        optimization_level (int): How much optimization to perform on the circuits.
            Higher levels generate more optimized circuits,
            at the expense of longer transpilation time.
            #. No optimization
            #. Light optimization
            #. Heavy optimization
            #. Highest optimization
            If None, level 1 will be chosen as default.

        pass_manager (PassManager): The pass manager to use during transpilation. If this
            arg is present, auto-selection of pass manager based on the transpile options
            will be turned off and this pass manager will be used directly.

        qobj_id (str): DEPRECATED: String identifier to annotate the Qobj.  This has no effect
            and the :attr:`~.QuantumCircuit.name` attribute of the input circuit(s) should be used
            instead.

        qobj_header (QobjHeader or dict): DEPRECATED: User input that will be inserted in Qobj
            header, and will also be copied to the corresponding :class:`qiskit.result.Result`
            header. Headers do not affect the run. Headers do not affect the run. This kwarg
            has no effect anymore and the :attr:`~.QuantumCircuit.metadata` attribute of the
            input circuit(s) should be used instead.

        shots (int): Number of repetitions of each circuit, for sampling. Default: 1024

        memory (bool): If True, per-shot measurement bitstrings are returned as well
            (provided the backend supports it). For OpenPulse jobs, only
            measurement level 2 supports this option. Default: False

        max_credits (int): DEPRECATED This parameter is deprecated as of Qiskit Terra 0.20.0
            and will be removed in a future release. This parameter has no effect on modern
            IBM Quantum systems, no alternative is necessary.

        seed_simulator (int): Random seed to control sampling, for when backend is a simulator

        default_qubit_los (Optional[List[float]]): List of job level qubit drive LO frequencies
            in Hz. Overridden by ``schedule_los`` if specified. Must have length ``n_qubits``.

        default_meas_los (Optional[List[float]]): List of job level measurement LO frequencies in
            Hz. Overridden by ``schedule_los`` if specified. Must have length ``n_qubits``.

        qubit_lo_range (Optional[List[List[float]]]): List of job level drive LO ranges each of form
            ``[range_min, range_max]`` in Hz. Used to validate ``qubit_lo_freq``. Must have length
            ``n_qubits``.

        meas_lo_range (Optional[List[List[float]]]): List of job level measurement LO ranges each of
            form ``[range_min, range_max]`` in Hz. Used to validate ``meas_lo_freq``. Must have
            length ``n_qubits``.

        schedule_los (list):
            Experiment level (ie circuit or schedule) LO frequency configurations for qubit drive
            and measurement channels. These values override the job level values from
            ``default_qubit_los`` and ``default_meas_los``. Frequencies are in Hz. Settable for qasm
            and pulse jobs.

            If a single LO config or dict is used, the values are set at job level. If a list is
            used, the list must be the size of the number of experiments in the job, except in the
            case of a single experiment. In this case, a frequency sweep will be assumed and one
            experiment will be created for every list entry.

            Not every channel is required to be specified. If not specified, the backend default
            value will be used.

        meas_level (int or MeasLevel): Set the appropriate level of the
            measurement output for pulse experiments.

        meas_return (str or MeasReturn): Level of measurement data for the
            backend to return For ``meas_level`` 0 and 1:
            ``"single"`` returns information from every shot.
            ``"avg"`` returns average measurement output (averaged over number
            of shots).

        memory_slots (int): Number of classical memory slots used in this job.

        memory_slot_size (int): Size of each memory slot if the output is Level 0.

        rep_time (int): Time per program execution in seconds. Must be from the list provided
            by the backend (``backend.configuration().rep_times``). Defaults to the first entry.

        rep_delay (float): Delay between programs in seconds. Only supported on certain
            backends (``backend.configuration().dynamic_reprate_enabled`` ). If supported,
            ``rep_delay`` will be used instead of ``rep_time`` and must be from the range supplied
            by the backend (``backend.configuration().rep_delay_range``). Default is given by
            ``backend.configuration().default_rep_delay``.

        parameter_binds (list[dict]): List of Parameter bindings over which the set of
            experiments will be executed. Each list element (bind) should be of the form
            ``{Parameter1: value1, Parameter2: value2, ...}``. All binds will be
            executed across all experiments, e.g. if parameter_binds is a
            length-:math:`n` list, and there are :math:`m` experiments, a total of :math:`m \\times n`
            experiments will be run (one for each experiment/bind pair).

        schedule_circuit (bool): If ``True``, ``experiments`` will be converted to
            :class:`qiskit.pulse.Schedule` objects prior to execution.

        inst_map (InstructionScheduleMap):
            Mapping of circuit operations to pulse schedules. If None, defaults to the
            ``instruction_schedule_map`` of ``backend``.

        meas_map (list(list(int))):
            List of sets of qubits that must be measured together. If None, defaults to
            the ``meas_map`` of ``backend``.

        scheduling_method (str or list(str)):
            Optionally specify a particular scheduling method.

        init_qubits (bool): Whether to reset the qubits to the ground state for each shot.
                            Default: ``True``.

        run_config (dict):
            Extra arguments used to configure the run (e.g. for Aer configurable backends).
            Refer to the backend documentation for details on these arguments.
            Note: for now, these keyword arguments will both be copied to the
            Qobj config, and passed to backend.run()

    Returns:
        Job: returns job instance derived from Job

    Raises:
        QiskitError: if the execution cannot be interpreted as either circuits or schedules

    Example:
        Construct a 5-qubit GHZ circuit and execute 4321 shots on a backend.

        .. jupyter-execute::

            from qiskit import QuantumCircuit, execute, BasicAer

            backend = BasicAer.get_backend('qasm_simulator')

            qc = QuantumCircuit(5, 5)
            qc.h(0)
            qc.cx(0, range(1, 5))
            qc.measure_all()

            job = execute(qc, backend, shots=4321)
    """
    if isinstance(experiments, (Schedule, ScheduleBlock)) or (
        isinstance(experiments, list) and isinstance(experiments[0], (Schedule, ScheduleBlock))
    ):
        # do not transpile a schedule circuit
        if schedule_circuit:
            raise QiskitError("Must supply QuantumCircuit to schedule circuit.")
    elif pass_manager is not None:
        # transpiling using pass_manager
        _check_conflicting_argument(
            optimization_level=optimization_level,
            basis_gates=basis_gates,
            coupling_map=coupling_map,
            seed_transpiler=seed_transpiler,
            backend_properties=backend_properties,
            initial_layout=initial_layout,
        )
        experiments = pass_manager.run(experiments)
    else:
        # transpiling the circuits using given transpile options
        experiments = transpile(
            experiments,
            basis_gates=basis_gates,
            coupling_map=coupling_map,
            backend_properties=backend_properties,
            initial_layout=initial_layout,
            seed_transpiler=seed_transpiler,
            optimization_level=optimization_level,
            backend=backend,
        )

    if schedule_circuit:
        experiments = schedule(
            circuits=experiments,
            backend=backend,
            inst_map=inst_map,
            meas_map=meas_map,
            method=scheduling_method,
        )
    if max_credits is not None:
        warnings.warn(
            "The `max_credits` parameter is deprecated as of Qiskit Terra 0.20.0, "
            "and will be removed in a future release. This parameter has no effect on "
            "modern IBM Quantum systems, and no alternative is necessary.",
            DeprecationWarning,
            stacklevel=2,
        )

    if qobj_id is not None:
        warnings.warn(
            "The qobj_id argument is deprecated as of the Qiskit Terra 0.21.0, "
            "and will be remvoed in a future release. This argument has no effect and "
            "is not used by any backends."
        )

    if qobj_header is not None:
        warnings.warn(
            "The qobj_header argument is deprecated as of the Qiskit Terra 0.21.0, "
            "and will be remvoed in a future release. This argument has no effect and "
            "is not used by any backends."
        )

    if isinstance(backend, Backend):
        start_time = time()
        run_kwargs = {
            "shots": shots,
            "memory": memory,
            "seed_simulator": seed_simulator,
            "qubit_lo_freq": default_qubit_los,
            "meas_lo_freq": default_meas_los,
            "qubit_lo_range": qubit_lo_range,
            "meas_lo_range": meas_lo_range,
            "schedule_los": schedule_los,
            "meas_level": meas_level,
            "meas_return": meas_return,
            "memory_slots": memory_slots,
            "memory_slot_size": memory_slot_size,
            "rep_time": rep_time,
            "rep_delay": rep_delay,
            "init_qubits": init_qubits,
        }
        for key in list(run_kwargs.keys()):
            if not hasattr(backend.options, key):
                if run_kwargs[key] is not None:
                    logger.info(
                        "%s backend doesn't support option %s so not passing that kwarg to run()",
                        backend.name,
                        key,
                    )
                del run_kwargs[key]
            elif run_kwargs[key] is None:
                del run_kwargs[key]

        if parameter_binds:
            run_kwargs["parameter_binds"] = parameter_binds
        run_kwargs.update(run_config)
        job = backend.run(experiments, **run_kwargs)
        end_time = time()
        _log_submission_time(start_time, end_time)
    else:
        raise QiskitError("Invalid backend type %s" % type(backend))
    return job


def _check_conflicting_argument(**kargs):
    conflicting_args = [arg for arg, value in kargs.items() if value]
    if conflicting_args:
        raise QiskitError(
            "The parameters pass_manager conflicts with the following "
            "parameter(s): {}.".format(", ".join(conflicting_args))
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Quantum information utility functions for states.
"""

import numpy as np

from qiskit.exceptions import QiskitError
from qiskit.quantum_info.states.statevector import Statevector
from qiskit.quantum_info.states.densitymatrix import DensityMatrix
from qiskit.quantum_info.operators.channel import SuperOp


def partial_trace(state, qargs):
    """Return reduced density matrix by tracing out part of quantum state.

    If all subsystems are traced over this returns the
    :meth:`~qiskit.quantum_info.DensityMatrix.trace` of the
    input state.

    Args:
        state (Statevector or DensityMatrix): the input state.
        qargs (list): The subsystems to trace over.

    Returns:
        DensityMatrix: The reduced density matrix.

    Raises:
        QiskitError: if input state is invalid.
    """
    state = _format_state(state, validate=False)

    # Compute traced shape
    traced_shape = state._op_shape.remove(qargs=qargs)
    # Convert vector shape to matrix shape
    traced_shape._dims_r = traced_shape._dims_l
    traced_shape._num_qargs_r = traced_shape._num_qargs_l

    # If we are tracing over all subsystems we return the trace
    if traced_shape.size == 0:
        return state.trace()

    # Statevector case
    if isinstance(state, Statevector):
        trace_systems = len(state._op_shape.dims_l()) - 1 - np.array(qargs)
        arr = state._data.reshape(state._op_shape.tensor_shape)
        rho = np.tensordot(arr, arr.conj(), axes=(trace_systems, trace_systems))
        rho = np.reshape(rho, traced_shape.shape)
        return DensityMatrix(rho, dims=traced_shape._dims_l)

    # Density matrix case
    # Empty partial trace case.
    if not qargs:
        return state.copy()
    # Trace first subsystem to avoid coping whole density matrix
    dims = state.dims(qargs)
    tr_op = SuperOp(np.eye(dims[0]).reshape(1, dims[0] ** 2), input_dims=[dims[0]], output_dims=[1])
    ret = state.evolve(tr_op, [qargs[0]])
    # Trace over remaining subsystems
    for qarg, dim in zip(qargs[1:], dims[1:]):
        tr_op = SuperOp(np.eye(dim).reshape(1, dim**2), input_dims=[dim], output_dims=[1])
        ret = ret.evolve(tr_op, [qarg])
    # Remove traced over subsystems which are listed as dimension 1
    ret._op_shape = traced_shape
    return ret


def shannon_entropy(pvec, base=2):
    r"""Compute the Shannon entropy of a probability vector.

    The shannon entropy of a probability vector
    :math:`\vec{p} = [p_0, ..., p_{n-1}]` is defined as

    .. math::

        H(\vec{p}) = \sum_{i=0}^{n-1} p_i \log_b(p_i)

    where :math:`b` is the log base and (default 2), and
    :math:`0 \log_b(0) \equiv 0`.

    Args:
        pvec (array_like): a probability vector.
        base (int): the base of the logarithm [Default: 2].

    Returns:
        float: The Shannon entropy H(pvec).
    """
    if base == 2:

        def logfn(x):
            return -x * np.log2(x)

    elif base == np.e:

        def logfn(x):
            return -x * np.log(x)

    else:

        def logfn(x):
            return -x * np.log(x) / np.log(base)

    h_val = 0.0
    for x in pvec:
        if 0 < x < 1:
            h_val += logfn(x)
    return h_val


def _format_state(state, validate=True):
    """Format input state into class object"""
    if isinstance(state, list):
        state = np.array(state, dtype=complex)
    if isinstance(state, np.ndarray):
        ndim = state.ndim
        if ndim == 1:
            state = Statevector(state)
        elif ndim == 2:
            dim1, dim2 = state.shape
            if dim2 == 1:
                state = Statevector(state)
            elif dim1 == dim2:
                state = DensityMatrix(state)
    if not isinstance(state, (Statevector, DensityMatrix)):
        raise QiskitError("Input is not a quantum state")
    if validate and not state.is_valid():
        raise QiskitError("Input quantum state is not a valid")
    return state


def _funm_svd(matrix, func):
    """Apply real scalar function to singular values of a matrix.

    Args:
        matrix (array_like): (N, N) Matrix at which to evaluate the function.
        func (callable): Callable object that evaluates a scalar function f.

    Returns:
        ndarray: funm (N, N) Value of the matrix function specified by func
                 evaluated at `A`.
    """
    import scipy.linalg as la

    unitary1, singular_values, unitary2 = la.svd(matrix)
    diag_func_singular = np.diag(func(singular_values))
    return unitary1.dot(diag_func_singular).dot(unitary2)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Gradient of probabilities with parameter shift
"""

from __future__ import annotations

from typing import Sequence

import numpy as np

from qiskit.algorithms import AlgorithmError
from qiskit.circuit import Parameter, QuantumCircuit
from qiskit.primitives import BaseSampler

from .base_sampler_gradient import BaseSamplerGradient
from .sampler_gradient_result import SamplerGradientResult
from .utils import _param_shift_preprocessing, _make_param_shift_parameter_values


class ParamShiftSamplerGradient(BaseSamplerGradient):
    """Compute the gradients of the sampling probability by the parameter shift rule."""

    def __init__(self, sampler: BaseSampler, **options):
        """
        Args:
            sampler: The sampler used to compute the gradients.
            options: Primitive backend runtime options used for circuit execution.
                The order of priority is: options in ``run`` method > gradient's
                default options > primitive's default setting.
                Higher priority setting overrides lower priority setting
        """
        self._gradient_circuits = {}
        super().__init__(sampler, **options)

    def _run(
        self,
        circuits: Sequence[QuantumCircuit],
        parameter_values: Sequence[Sequence[float]],
        parameters: Sequence[Sequence[Parameter] | None],
        **options,
    ) -> SamplerGradientResult:
        """Compute the sampler gradients on the given circuits."""
        jobs, result_indices_all, coeffs_all, metadata_ = [], [], [], []
        for circuit, parameter_values_, parameters_ in zip(circuits, parameter_values, parameters):
            # a set of parameters to be differentiated
            if parameters_ is None:
                param_set = set(circuit.parameters)
            else:
                param_set = set(parameters_)
            metadata_.append({"parameters": [p for p in circuit.parameters if p in param_set]})

            if self._gradient_circuits.get(id(circuit)):
                gradient_circuit, base_parameter_values_all = self._gradient_circuits[id(circuit)]
            else:
                gradient_circuit, base_parameter_values_all = _param_shift_preprocessing(circuit)
                self._gradient_circuits[id(circuit)] = (
                    gradient_circuit,
                    base_parameter_values_all,
                )

            (
                gradient_parameter_values_plus,
                gradient_parameter_values_minus,
                result_indices,
                coeffs,
            ) = _make_param_shift_parameter_values(
                gradient_circuit_data=gradient_circuit,
                base_parameter_values=base_parameter_values_all,
                parameter_values=parameter_values_,
                param_set=param_set,
            )
            n = 2 * len(gradient_parameter_values_plus)

            job = self._sampler.run(
                [gradient_circuit.gradient_circuit] * n,
                gradient_parameter_values_plus + gradient_parameter_values_minus,
                **options,
            )
            jobs.append(job)
            result_indices_all.append(result_indices)
            coeffs_all.append(coeffs)

        # combine the results
        try:
            results = [job.result() for job in jobs]
        except Exception as exc:
            raise AlgorithmError("Sampler job failed.") from exc

        gradients = []
        for i, result in enumerate(results):
            n = len(result.quasi_dists) // 2
            grad_dists = np.zeros((len(metadata_[i]["parameters"]), 2 ** circuits[i].num_qubits))
            for idx, coeff, dist_plus, dist_minus in zip(
                result_indices_all[i], coeffs_all[i], result.quasi_dists[:n], result.quasi_dists[n:]
            ):
                grad_dists[idx][list(dist_plus.keys())] += (
                    np.array(list(dist_plus.values())) * coeff
                )
                grad_dists[idx][list(dist_minus.keys())] -= (
                    np.array(list(dist_minus.values())) * coeff
                )

            gradient_ = []
            for grad_dist in grad_dists:
                gradient_.append(dict(enumerate(grad_dist)))
            gradients.append(gradient_)

        opt = self._get_local_options(options)
        return SamplerGradientResult(gradients=gradients, metadata=metadata_, options=opt)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Result object for p-VQD."""
from __future__ import annotations

import numpy as np

from qiskit.circuit import QuantumCircuit
from ..time_evolution_result import TimeEvolutionResult


class PVQDResult(TimeEvolutionResult):
    """The result object for the p-VQD algorithm."""

    def __init__(
        self,
        evolved_state: QuantumCircuit,
        aux_ops_evaluated: list[tuple[complex, complex]] | None = None,
        times: list[float] | None = None,
        parameters: list[np.ndarray] | None = None,
        fidelities: list[float] | None = None,
        estimated_error: float | None = None,
        observables: list[list[float]] | None = None,
    ):
        """
        Args:
            evolved_state: An evolved quantum state.
            aux_ops_evaluated: Optional list of observables for which expected values on an evolved
                state are calculated. These values are in fact tuples formatted as (mean, standard
                deviation).
            times: The times evaluated during the time integration.
            parameters: The parameter values at each evaluation time.
            fidelities: The fidelity of the Trotter step and variational update at each iteration.
            estimated_error: The overall estimated error evaluated as one minus the
                product of all fidelities.
            observables: The value of the observables evaluated at each iteration.
        """
        super().__init__(evolved_state, aux_ops_evaluated)
        self.times = times
        self.parameters = parameters
        self.fidelities = fidelities
        self.estimated_error = estimated_error
        self.observables = observables



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Collect sequences of uninterrupted gates acting on 2 qubits."""

from collections import defaultdict

from qiskit.transpiler.basepasses import AnalysisPass


class Collect2qBlocks(AnalysisPass):
    """Collect two-qubit subcircuits."""

    def run(self, dag):
        """Run the Collect2qBlocks pass on `dag`.

        The blocks contain "op" nodes in topological order such that all gates
        in a block act on the same qubits and are adjacent in the circuit.

        After the execution, ``property_set['block_list']`` is set to a list of
        tuples of "op" node.
        """
        self.property_set["commutation_set"] = defaultdict(list)
        self.property_set["block_list"] = dag.collect_2q_runs()

        return dag



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Decompose an arbitrary 2*2 unitary into three rotation gates: U=R_zR_yR_z.

Note that the decomposition is up to a global phase shift.
(This is a well known decomposition, which can be found for example in Nielsen and Chuang's book
"Quantum computation and quantum information".)
"""

import cmath

import numpy as np

from qiskit.circuit import QuantumRegister, Qubit, QuantumCircuit
from qiskit.circuit.gate import Gate
from qiskit.circuit.exceptions import CircuitError
from qiskit.quantum_info.operators.predicates import is_unitary_matrix
from qiskit.exceptions import QiskitError

_EPS = 1e-10  # global variable used to chop very small numbers to zero


class SingleQubitUnitary(Gate):
    """
    u = 2*2 unitary (given as a (complex) numpy.ndarray)

    mode - determines the used decomposition by providing the rotation axes

    up_to_diagonal - the single-qubit unitary is decomposed up to a diagonal matrix,
                     i.e. a unitary u' is implemented such that there exists a 2*2 diagonal
                     gate d with u = d.dot(u').
    """

    def __init__(self, unitary_matrix, mode="ZYZ", up_to_diagonal=False):
        """Create a new single qubit gate based on the unitary ``u``."""
        if mode not in ["ZYZ"]:
            raise QiskitError("The decomposition mode is not known.")
        # Check if the matrix u has the right dimensions and if it is a unitary
        if unitary_matrix.shape != (2, 2):
            raise QiskitError("The dimension of the input matrix is not equal to (2,2).")
        if not is_unitary_matrix(unitary_matrix):
            raise QiskitError("The 2*2 matrix is not unitary.")

        self.mode = mode
        self.up_to_diagonal = up_to_diagonal
        self._diag = None

        # Create new gate
        super().__init__("squ", 1, [unitary_matrix])

    def inverse(self):
        """Return the inverse.

        Note that the resulting gate has an empty ``params`` property.
        """
        inverse_gate = Gate(
            name=self.name + "_dg", num_qubits=self.num_qubits, params=[]
        )  # removing the params because arrays are deprecated

        definition = QuantumCircuit(*self.definition.qregs)
        for inst in reversed(self._definition):
            definition._append(inst.replace(operation=inst.operation.inverse()))
        inverse_gate.definition = definition
        return inverse_gate

    @property
    def diag(self):
        """Returns the diagonal gate D up to which the single-qubit unitary u is implemented.

        I.e. u=D.u', where u' is the unitary implemented by the found circuit.
        """
        if self._diag is None:
            self._define()
        return self._diag

    def _define(self):
        """Define the gate using the decomposition."""

        if self.mode == "ZYZ":
            circuit, diag = self._zyz_circuit()
        else:
            raise QiskitError("The decomposition mode is not known.")

        self._diag = diag

        self.definition = circuit

    def _zyz_circuit(self):
        """Get the circuit for the ZYZ decomposition."""
        q = QuantumRegister(self.num_qubits)
        qc = QuantumCircuit(q, name=self.name)

        diag = [1.0, 1.0]
        alpha, beta, gamma, _ = self._zyz_dec()

        if abs(alpha) > _EPS:
            qc.rz(alpha, q[0])
        if abs(beta) > _EPS:
            qc.ry(beta, q[0])
        if abs(gamma) > _EPS:
            if self.up_to_diagonal:
                diag = [np.exp(-1j * gamma / 2.0), np.exp(1j * gamma / 2.0)]
            else:
                qc.rz(gamma, q[0])

        return qc, diag

    def _zyz_dec(self):
        """Finds rotation angles (a,b,c,d) in the decomposition u=exp(id)*Rz(c).Ry(b).Rz(a).

        Note that where "." denotes matrix multiplication.
        """
        unitary = self.params[0]
        u00 = unitary.item(0, 0)
        u01 = unitary.item(0, 1)
        u10 = unitary.item(1, 0)
        u11 = unitary.item(1, 1)
        # Handle special case if the entry (0,0) of the unitary is equal to zero
        if np.abs(u00) < _EPS:
            # Note that u10 can't be zero, since u is unitary (and u00 == 0)
            gamma = cmath.phase(-u01 / u10)
            delta = cmath.phase(u01 * np.exp(-1j * gamma / 2))
            return 0.0, -np.pi, -gamma, delta
        # Handle special case if the entry (0,1) of the unitary is equal to zero
        if np.abs(u01) < _EPS:
            # Note that u11 can't be zero, since u is unitary (and u01 == 0)
            gamma = cmath.phase(u00 / u11)
            delta = cmath.phase(u00 * np.exp(-1j * gamma / 2))
            return 0.0, 0.0, -gamma, delta
        beta = 2 * np.arccos(np.abs(u00))
        if np.sin(beta / 2) - np.cos(beta / 2) > 0:
            gamma = cmath.phase(-u00 / u10)
            alpha = cmath.phase(u00 / u01)
        else:
            gamma = -cmath.phase(-u10 / u00)
            alpha = -cmath.phase(u01 / u00)
        delta = cmath.phase(u00 * np.exp(-1j * (alpha + gamma) / 2))
        # The decomposition works with another convention for the rotation gates
        # (the one using negative angles).
        # Therefore, we have to take the inverse of the angles at the end.
        return -alpha, -beta, -gamma, delta

    def validate_parameter(self, parameter):
        """Single-qubit unitary gate parameter has to be an ndarray."""
        if isinstance(parameter, np.ndarray):
            return parameter
        else:
            raise CircuitError(f"invalid param type {type(parameter)} in gate {self.name}")


def squ(
    self,
    unitary_matrix,
    qubit,
    mode="ZYZ",
    up_to_diagonal=False,
):
    """Decompose an arbitrary 2*2 unitary into three rotation gates.

    Note that the decomposition is up to a global phase shift.
    (This is a well known decomposition, which can be found for example in Nielsen and Chuang's book
    "Quantum computation and quantum information".)

    Args:
        unitary_matrix (ndarray): 2*2 unitary (given as a (complex) ndarray).
        qubit (QuantumRegister or Qubit): The qubit which the gate is acting on.
        mode (string): determines the used decomposition by providing the rotation axes.
            The allowed modes are: "ZYZ" (default)
        up_to_diagonal (bool):  if set to True, the single-qubit unitary is decomposed up to
            a diagonal matrix, i.e. a unitary u' is implemented such that there exists a 2*2
            diagonal gate d with u = d.dot(u')

    Returns:
        InstructionSet: The single-qubit unitary instruction attached to the circuit.

    Raises:
        QiskitError: if the format is wrong; if the array u is not unitary
    """

    if isinstance(qubit, QuantumRegister):
        qubit = qubit[:]
        if len(qubit) == 1:
            qubit = qubit[0]
        else:
            raise QiskitError(
                "The target qubit is a QuantumRegister containing more than one qubit."
            )
    # Check if there is one target qubit provided
    if not isinstance(qubit, Qubit):
        raise QiskitError("The target qubit is not a single qubit from a QuantumRegister.")
    return self.append(SingleQubitUnitary(unitary_matrix, mode, up_to_diagonal), [qubit], [])


QuantumCircuit.squ = squ
